"""This module defines a generated parser model."""

from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.HexStringModelElement import HexStringModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement


def get_model():
    """Return a model to parse Suricata Event logs from the AIT-LDS."""
    conn = SequenceModelElement("conn", [
                FixedDataModelElement("src_ip_str", b'"src_ip":"'),
                FirstMatchModelElement("ip", [
                    SequenceModelElement("ipv4", [
                        IpAddressDataModelElement("src_ip"),
                        FixedDataModelElement("src_port_str", b'","src_port":'),
                        DecimalIntegerValueModelElement("src_port"),
                        FixedDataModelElement("dest_ip_str", b',"dest_ip":"'),
                        IpAddressDataModelElement("dest_ip"),
                        FixedDataModelElement("dest_port_str", b'","dest_port":'),
                        DecimalIntegerValueModelElement("dest_port"),
                        FixedDataModelElement("proto_str", b',"proto":"'),
                        DelimitedDataModelElement("proto", b'"'),
                        FixedDataModelElement("quote", b'"')
                        ]),
                    SequenceModelElement("ipv6", [
                        DelimitedDataModelElement("src_ip", b'"'),
                        FixedDataModelElement("dest_ip_str", b'","dest_ip":"'),
                        DelimitedDataModelElement("dest_ip", b'"'),
                        FixedDataModelElement("proto_str", b'","proto":"'),
                        DelimitedDataModelElement("proto", b'"'),
                        FixedDataModelElement("icmp_type_str", b'","icmp_type":'),
                        DecimalIntegerValueModelElement("icmp_type"),
                        FixedDataModelElement("icmp_code_str", b',"icmp_code":'),
                        DecimalIntegerValueModelElement("icmp_code"),
                        ]),
                    ])
        ])

    http = SequenceModelElement("http", [
                FixedDataModelElement("hostname_str", b',"http":{"hostname":"'),
                DelimitedDataModelElement("hostname", b'"'),
                FixedDataModelElement("url_str", b'","url":"'),
                DelimitedDataModelElement("url", b'"', escape=b"\\"),
                FixedDataModelElement("http_user_agent_str", b'","http_user_agent":"'),
                DelimitedDataModelElement("http_user_agent", b'"'),
                OptionalMatchModelElement(
                    "content_type", SequenceModelElement("content_type", [
                        FixedDataModelElement("http_content_type_str", b'","http_content_type":"'),
                        DelimitedDataModelElement("http_content_type", b'"'),
                        ])),
                OptionalMatchModelElement(
                    "http_refer", SequenceModelElement("http_refer", [
                        FixedDataModelElement("http_refer_str", b'","http_refer":"'),
                        DelimitedDataModelElement("http_refer", b'"'),
                        ])),
                FixedDataModelElement("http_method_str", b'","http_method":"'),
                DelimitedDataModelElement("http_method", b'"'),
                FixedDataModelElement("protocol_str", b'","protocol":"'),
                DelimitedDataModelElement("protocol", b'"'),
                FixedDataModelElement("quote_str", b'"'),
                OptionalMatchModelElement(
                    "status", SequenceModelElement("status", [
                        FixedDataModelElement("status_str", b',"status":'),
                        DecimalIntegerValueModelElement("status"),
                        ])),
                OptionalMatchModelElement(
                    "redirect", SequenceModelElement("redirect", [
                        FixedDataModelElement("redirect_str", b',"redirect":"'),
                        DelimitedDataModelElement("redirect", b'"'),
                        FixedDataModelElement("quote_str", b'"')
                        ])),
                FixedDataModelElement("length_str", b',"length":'),
                DecimalIntegerValueModelElement("length"),
                FixedDataModelElement("brack_str", b"}")
        ])

    model = SequenceModelElement("model", [
        FixedDataModelElement("time_str", b'{"timestamp":"'),
        DateTimeModelElement("time", b"%Y-%m-%dT%H:%M:%S.%f%z"),
        FixedDataModelElement("comma_str", b'",'),
        OptionalMatchModelElement(
            "flow_id", SequenceModelElement("flow_id", [
                FixedDataModelElement("flow_id_str", b'"flow_id":'),
                DecimalIntegerValueModelElement("flow_id"),
                FixedDataModelElement("comma_str", b",")])),
        OptionalMatchModelElement(
            "in_iface", SequenceModelElement("in_iface", [
                FixedDataModelElement("in_iface_str", b'"in_iface":"'),
                DelimitedDataModelElement("in_iface", b'"'),
                FixedDataModelElement("comma_str", b'",')])),
        FixedDataModelElement("event_type_str", b'"event_type":"'),
        FirstMatchModelElement("event_type", [
            SequenceModelElement("dns", [
                FixedDataModelElement("dns_str", b'dns",'),
                conn,
                SequenceModelElement("dns", [
                    FixedDataModelElement("type_str", b',"dns":{"type":"'),
                    DelimitedDataModelElement("type", b'"'),
                    FixedDataModelElement("id_str", b'","id":'),
                    DecimalIntegerValueModelElement("id"),
                    OptionalMatchModelElement(
                        "rcode", SequenceModelElement("rcode", [
                            FixedDataModelElement("rcode_str", b',"rcode":"'),
                            DelimitedDataModelElement("rcode", b'"'),
                            FixedDataModelElement("quote_str", b'"')])),
                    FixedDataModelElement("rrname_str", b',"rrname":"'),
                    DelimitedDataModelElement("rrname", b'"'),
                    OptionalMatchModelElement("rrtype", SequenceModelElement("rrtype", [
                            FixedDataModelElement("rrtype_str", b'","rrtype":"'),
                            DelimitedDataModelElement("rrtype", b'"')])),
                    FixedDataModelElement("quote", b'"'),
                    OptionalMatchModelElement(
                        "tx_id", SequenceModelElement("tx_id", [
                            FixedDataModelElement("tx_id_str", b',"tx_id":'),
                            DecimalIntegerValueModelElement("tx_id")])),
                    OptionalMatchModelElement("ttl", SequenceModelElement("ttl", [
                            FixedDataModelElement("ttl_str", b',"ttl":'),
                            DecimalIntegerValueModelElement("ttl")])),
                    OptionalMatchModelElement(
                        "rdata", SequenceModelElement("rdata", [
                            FixedDataModelElement("rdata_str", b',"rdata":"'),
                            DelimitedDataModelElement("rdata", b'"'),
                            FixedDataModelElement("quote_str", b'"')])),
                    FixedDataModelElement("brack_str", b"}}")
                    ]),
                ]),
            SequenceModelElement("flow", [
                FixedDataModelElement("flow_str", b'flow",'),
                conn,
                OptionalMatchModelElement(
                    "app_proto", SequenceModelElement("app_proto", [
                        FixedDataModelElement("app_proto_str", b',"app_proto":"'),
                        DelimitedDataModelElement("app_proto", b'"'),
                        FixedDataModelElement("quote_str", b'"')
                        ])
                    ),
                OptionalMatchModelElement(
                    "app_proto_tc", SequenceModelElement("app_proto_tc", [
                        FixedDataModelElement("app_proto_tc_str", b',"app_proto_tc":"'),
                        DelimitedDataModelElement("app_proto_tc", b'"'),
                        FixedDataModelElement("quote_str", b'"')
                        ])
                    ),
                SequenceModelElement("flow", [
                    FixedDataModelElement("pkts_toserver_str", b',"flow":{"pkts_toserver":'),
                    DecimalIntegerValueModelElement("pkts_toserver"),
                    FixedDataModelElement("pkts_toclient_str", b',"pkts_toclient":'),
                    DecimalIntegerValueModelElement("pkts_toclient"),
                    FixedDataModelElement("bytes_toserver_str", b',"bytes_toserver":'),
                    DecimalIntegerValueModelElement("bytes_toserver"),
                    FixedDataModelElement("bytes_toclient_str", b',"bytes_toclient":'),
                    DecimalIntegerValueModelElement("bytes_toclient"),
                    FixedDataModelElement("start_str", b',"start":"'),
                    DelimitedDataModelElement("start", b'"'),
                    FixedDataModelElement("end_str", b'","end":"'),
                    DelimitedDataModelElement("end", b'"'),
                    FixedDataModelElement("age_str", b'","age":'),
                    DecimalIntegerValueModelElement("age"),
                    FixedDataModelElement("state_str", b',"state":"'),
                    DelimitedDataModelElement("state", b'"'),
                    FixedDataModelElement("reason_str", b'","reason":"'),
                    DelimitedDataModelElement("reason", b'"'),
                    FixedDataModelElement("alerted_str", b'","alerted":'),
                    FixedWordlistDataModelElement("alerted", [b"true", b"false"]),
                    FixedDataModelElement("brack_str1", b"}"),
                    OptionalMatchModelElement(
                        "tcp", SequenceModelElement("tcp", [
                            FixedDataModelElement("tcp_flags_str", b',"tcp":{"tcp_flags":"'),
                            HexStringModelElement("tcp_flags"),
                            FixedDataModelElement("tcp_flags_ts_str", b'","tcp_flags_ts":"'),
                            HexStringModelElement("tcp_flags_ts"),
                            FixedDataModelElement("tcp_flags_tc_str", b'","tcp_flags_tc":"'),
                            HexStringModelElement("tcp_flags_tc"),
                            OptionalMatchModelElement(
                                "flags", SequenceModelElement("flags", [
                                    FixedDataModelElement("syn_str", b'","syn":'),
                                    FixedWordlistDataModelElement("syn", [b"true", b"false"]),
                                    OptionalMatchModelElement(
                                        "fin", SequenceModelElement("fin", [
                                            FixedDataModelElement("fin_str", b',"fin":'),
                                            FixedWordlistDataModelElement("fin", [b"true", b"false"]),
                                            ])
                                        ),
                                    OptionalMatchModelElement(
                                        "rst", SequenceModelElement("rst", [
                                            FixedDataModelElement("rst_str", b',"rst":'),
                                            FixedWordlistDataModelElement("rst", [b"true", b"false"]),
                                            ])
                                        ),
                                    OptionalMatchModelElement(
                                        "psh", SequenceModelElement("psh", [
                                            FixedDataModelElement("psh_str", b',"psh":'),
                                            FixedWordlistDataModelElement("psh", [b"true", b"false"]),
                                            ])
                                        ),
                                    FixedDataModelElement("ack_str", b',"ack":'),
                                    FixedWordlistDataModelElement("ack", [b"true", b"false"]),
                                    FixedDataModelElement("tcp_state_str", b',"state":"'),
                                    DelimitedDataModelElement("tcp_state", b'"'),
                                    ])
                                ),
                            FixedDataModelElement("tcp_brack_str", b'"}'),
                            ])
                        ),
                    FixedDataModelElement("brack_str2", b"}")
                    ]),
                ]),
            SequenceModelElement("http", [
                FixedDataModelElement("http_str", b'http",'),
                conn,
                FixedDataModelElement("tx_id_str", b',"tx_id":'),
                DecimalIntegerValueModelElement("tx_id"),
                http,
                FixedDataModelElement("brack_str", b"}")
                ]),
            SequenceModelElement("fileinfo", [
                FixedDataModelElement("fileinfo_str", b'fileinfo",'),
                conn,
                http,
                FixedDataModelElement("app_proto_str", b',"app_proto":"'),
                DelimitedDataModelElement("app_proto", b'"'),
                SequenceModelElement("fileinfo", [
                    FixedDataModelElement("fileinfo_str", b'","fileinfo":{'),
                    OptionalMatchModelElement(
                        "filename", SequenceModelElement("filename", [
                            FixedDataModelElement("filename_str", b'"filename":"'),
                            DelimitedDataModelElement("filename", b'"'),
                            FixedDataModelElement("quote_str", b'",')
                        ])
                    ),
                    FixedDataModelElement("state_str", b'"state":"'),
                    DelimitedDataModelElement("state", b'"'),
                    FixedDataModelElement("stored_str", b'","stored":'),
                    FixedWordlistDataModelElement("stored", [b"true", b"false"]),
                    FixedDataModelElement("size_str", b',"size":'),
                    DecimalIntegerValueModelElement("size"),
                    FixedDataModelElement("tx_id_str", b',"tx_id":'),
                    DecimalIntegerValueModelElement("tx_id"),
                    FixedDataModelElement("brack_str", b"}}")
                    ]),
                ]),
            SequenceModelElement("stats", [
                FixedDataModelElement("stats_str", b'stats",'),
                FixedDataModelElement("uptime_str", b'"stats":{"uptime":'),
                DecimalIntegerValueModelElement("uptime"),
                SequenceModelElement("capture", [
                    FixedDataModelElement("capture_str", b',"capture":{'),
                    FixedDataModelElement("kernel_packets_str", b'"kernel_packets":'),
                    DecimalIntegerValueModelElement("kernel_packets"),
                    FixedDataModelElement("kernel_drops_str", b',"kernel_drops":'),
                    DecimalIntegerValueModelElement("kernel_drops"),
                    FixedDataModelElement("brack_str", b"}")
                    ]),
                SequenceModelElement("decoder", [
                    FixedDataModelElement("pkts_str", b',"decoder":{"pkts":'),
                    DecimalIntegerValueModelElement("pkts"),
                    FixedDataModelElement("bytes_str", b',"bytes":'),
                    DecimalIntegerValueModelElement("bytes"),
                    FixedDataModelElement("invalid_str", b',"invalid":'),
                    DecimalIntegerValueModelElement("invalid"),
                    FixedDataModelElement("ipv4_str", b',"ipv4":'),
                    DecimalIntegerValueModelElement("ipv4"),
                    FixedDataModelElement("ipv6_str", b',"ipv6":'),
                    DecimalIntegerValueModelElement("ipv6"),
                    FixedDataModelElement("ethernet_str", b',"ethernet":'),
                    DecimalIntegerValueModelElement("ethernet"),
                    FixedDataModelElement("raw_str", b',"raw":'),
                    DecimalIntegerValueModelElement("raw"),
                    FixedDataModelElement("null_str", b',"null":'),
                    DecimalIntegerValueModelElement("null"),
                    FixedDataModelElement("sll_str", b',"sll":'),
                    DecimalIntegerValueModelElement("sll"),
                    FixedDataModelElement("tcp_str", b',"tcp":'),
                    DecimalIntegerValueModelElement("tcp"),
                    FixedDataModelElement("udp_str", b',"udp":'),
                    DecimalIntegerValueModelElement("udp"),
                    FixedDataModelElement("sctp_str", b',"sctp":'),
                    DecimalIntegerValueModelElement("sctp"),
                    FixedDataModelElement("icmpv4_str", b',"icmpv4":'),
                    DecimalIntegerValueModelElement("icmpv4"),
                    FixedDataModelElement("icmpv6_str", b',"icmpv6":'),
                    DecimalIntegerValueModelElement("icmpv6"),
                    FixedDataModelElement("ppp_str", b',"ppp":'),
                    DecimalIntegerValueModelElement("ppp"),
                    FixedDataModelElement("pppoe_str", b',"pppoe":'),
                    DecimalIntegerValueModelElement("pppoe"),
                    FixedDataModelElement("gre_str", b',"gre":'),
                    DecimalIntegerValueModelElement("gre"),
                    FixedDataModelElement("vlan_str", b',"vlan":'),
                    DecimalIntegerValueModelElement("vlan"),
                    FixedDataModelElement("vlan_qinq_str", b',"vlan_qinq":'),
                    DecimalIntegerValueModelElement("vlan_qinq"),
                    FixedDataModelElement("teredo_str", b',"teredo":'),
                    DecimalIntegerValueModelElement("teredo"),
                    FixedDataModelElement("ipv4_in_ipv6_str", b',"ipv4_in_ipv6":'),
                    DecimalIntegerValueModelElement("ipv4_in_ipv6"),
                    FixedDataModelElement("ipv6_in_ipv6_str", b',"ipv6_in_ipv6":'),
                    DecimalIntegerValueModelElement("ipv6_in_ipv6"),
                    FixedDataModelElement("mpls_str", b',"mpls":'),
                    DecimalIntegerValueModelElement("mpls"),
                    FixedDataModelElement("avg_pkt_size_str", b',"avg_pkt_size":'),
                    DecimalIntegerValueModelElement("avg_pkt_size"),
                    FixedDataModelElement("max_pkt_size_str", b',"max_pkt_size":'),
                    DecimalIntegerValueModelElement("max_pkt_size"),
                    FixedDataModelElement("erspan_str", b',"erspan":'),
                    DecimalIntegerValueModelElement("erspan"),
                    SequenceModelElement("ipraw", [
                        FixedDataModelElement("invalid_ip_version_str", b',"ipraw":{"invalid_ip_version":'),
                        DecimalIntegerValueModelElement("invalid_ip_version"),
                        ]),
                    SequenceModelElement("ltnull", [
                        FixedDataModelElement("ipraw_pkt_too_small_str", b'},"ltnull":{"pkt_too_small":'),
                        DecimalIntegerValueModelElement("ipraw_pkt_too_small"),
                        FixedDataModelElement("unsupported_type", b',"unsupported_type":'),
                        DecimalIntegerValueModelElement("unsupported_type"),
                        ]),
                    SequenceModelElement("dce", [
                        FixedDataModelElement("dce_pkt_too_small_str", b'},"dce":{"pkt_too_small":'),
                        DecimalIntegerValueModelElement("dce_pkt_too_small"),
                        FixedDataModelElement("brack_str", b"}")
                    ])
                ]),
                SequenceModelElement("flow", [
                    FixedDataModelElement("memcap_str", b'},"flow":{"memcap":'),
                    DecimalIntegerValueModelElement("memcap"),
                    FixedDataModelElement("spare_str", b',"spare":'),
                    DecimalIntegerValueModelElement("spare"),
                    FixedDataModelElement("emerg_mode_entered_str", b',"emerg_mode_entered":'),
                    DecimalIntegerValueModelElement("emerg_mode_entered"),
                    FixedDataModelElement("emerg_mode_over_str", b',"emerg_mode_over":'),
                    DecimalIntegerValueModelElement("emerg_mode_over"),
                    FixedDataModelElement("tcp_reuse_str", b',"tcp_reuse":'),
                    DecimalIntegerValueModelElement("tcp_reuse"),
                    FixedDataModelElement("memuse_str", b',"memuse":'),
                    DecimalIntegerValueModelElement("memuse"),
                ]),
                SequenceModelElement("defrag", [
                    SequenceModelElement("ipv4", [
                        FixedDataModelElement("fragments_str", b'},"defrag":{"ipv4":{"fragments":'),
                        DecimalIntegerValueModelElement("fragments"),
                        FixedDataModelElement("reassembled_str", b',"reassembled":'),
                        DecimalIntegerValueModelElement("reassembled_str"),
                        FixedDataModelElement("timeouts_str", b',"timeouts":'),
                        DecimalIntegerValueModelElement("timeouts"),
                    ]),
                    SequenceModelElement("ipv6", [
                        FixedDataModelElement("fragments_str", b'},"ipv6":{"fragments":'),
                        DecimalIntegerValueModelElement("fragments"),
                        FixedDataModelElement("reassembled_str", b',"reassembled":'),
                        DecimalIntegerValueModelElement("reassembled_str"),
                        FixedDataModelElement("timeouts_str", b',"timeouts":'),
                        DecimalIntegerValueModelElement("timeouts"),
                    ]),
                    FixedDataModelElement("max_frag_hits_str", b'},"max_frag_hits":'),
                    DecimalIntegerValueModelElement("max_frag_hits"),
                ]),
                SequenceModelElement("tcp", [
                    FixedDataModelElement("sessions_str", b'},"tcp":{"sessions":'),
                    DecimalIntegerValueModelElement("sessions"),
                    FixedDataModelElement("ssn_memcap_drop_str", b',"ssn_memcap_drop":'),
                    DecimalIntegerValueModelElement("ssn_memcap_drop"),
                    FixedDataModelElement("pseudo_str", b',"pseudo":'),
                    DecimalIntegerValueModelElement("pseudo"),
                    FixedDataModelElement("pseudo_failed_str", b',"pseudo_failed":'),
                    DecimalIntegerValueModelElement("pseudo_failed"),
                    FixedDataModelElement("invalid_checksum_str", b',"invalid_checksum":'),
                    DecimalIntegerValueModelElement("invalid_checksum"),
                    FixedDataModelElement("no_flow_str", b',"no_flow":'),
                    DecimalIntegerValueModelElement("no_flow"),
                    FixedDataModelElement("syn_str", b',"syn":'),
                    DecimalIntegerValueModelElement("syn"),
                    FixedDataModelElement("synack_str", b',"synack":'),
                    DecimalIntegerValueModelElement("synack"),
                    FixedDataModelElement("rst_str", b',"rst":'),
                    DecimalIntegerValueModelElement("rst"),
                    FixedDataModelElement("segment_memcap_drop_str", b',"segment_memcap_drop":'),
                    DecimalIntegerValueModelElement("segment_memcap_drop"),
                    FixedDataModelElement("stream_depth_reached_str", b',"stream_depth_reached":'),
                    DecimalIntegerValueModelElement("stream_depth_reached"),
                    FixedDataModelElement("reassembly_gap_str", b',"reassembly_gap":'),
                    DecimalIntegerValueModelElement("reassembly_gap"),
                    FixedDataModelElement("memuse_str", b',"memuse":'),
                    DecimalIntegerValueModelElement("memuse"),
                    FixedDataModelElement("reassembly_memuse_str", b',"reassembly_memuse":'),
                    DecimalIntegerValueModelElement("reassembly_memuse"),
                    ]),
                SequenceModelElement("detect", [
                    FixedDataModelElement("alert_str", b'},"detect":{"alert":'),
                    DecimalIntegerValueModelElement("alert")
                    ]),
                SequenceModelElement("app_layer", [
                    SequenceModelElement("flow", [
                        FixedDataModelElement("http_str", b'},"app_layer":{"flow":{"http":'),
                        DecimalIntegerValueModelElement("http"),
                        FixedDataModelElement("ftp_str", b',"ftp":'),
                        DecimalIntegerValueModelElement("ftp"),
                        FixedDataModelElement("smtp_str", b',"smtp":'),
                        DecimalIntegerValueModelElement("smtp"),
                        FixedDataModelElement("tls_str", b',"tls":'),
                        DecimalIntegerValueModelElement("tls"),
                        FixedDataModelElement("ssh_str", b',"ssh":'),
                        DecimalIntegerValueModelElement("ssh"),
                        FixedDataModelElement("imap_str", b',"imap":'),
                        DecimalIntegerValueModelElement("imap"),
                        FixedDataModelElement("msn_str", b',"msn":'),
                        DecimalIntegerValueModelElement("msn"),
                        FixedDataModelElement("smb_str", b',"smb":'),
                        DecimalIntegerValueModelElement("smb"),
                        FixedDataModelElement("dcerpc_tcp_str", b',"dcerpc_tcp":'),
                        DecimalIntegerValueModelElement("dcerpc_tcp"),
                        FixedDataModelElement("dns_tcp_str", b',"dns_tcp":'),
                        DecimalIntegerValueModelElement("dns_tcp"),
                        FixedDataModelElement("failed_tcp_str", b',"failed_tcp":'),
                        DecimalIntegerValueModelElement("failed_tcp"),
                        FixedDataModelElement("dcerpc_udp_str", b',"dcerpc_udp":'),
                        DecimalIntegerValueModelElement("dcerpc_udp"),
                        FixedDataModelElement("dns_udp_str", b',"dns_udp":'),
                        DecimalIntegerValueModelElement("dns_udp"),
                        FixedDataModelElement("failed_udp_str", b',"failed_udp":'),
                        DecimalIntegerValueModelElement("failed_udp"),
                        ]),
                    SequenceModelElement("tx", [
                            FixedDataModelElement("http_str", b'},"tx":{"http":'),
                            DecimalIntegerValueModelElement("http"),
                            FixedDataModelElement("smtp_str", b',"smtp":'),
                            DecimalIntegerValueModelElement("smtp"),
                            FixedDataModelElement("tls_str", b',"tls":'),
                            DecimalIntegerValueModelElement("tls"),
                            FixedDataModelElement("dns_tcp_str", b',"dns_tcp":'),
                            DecimalIntegerValueModelElement("dns_tcp"),
                            FixedDataModelElement("dns_udp_str", b',"dns_udp":'),
                            DecimalIntegerValueModelElement("dns_udp"),
                        ])
                    ]),
                SequenceModelElement("flow_mgr", [
                    FixedDataModelElement("closed_pruned_str", b'}},"flow_mgr":{"closed_pruned":'),
                    DecimalIntegerValueModelElement("closed_pruned"),
                    FixedDataModelElement("new_pruned_str", b',"new_pruned":'),
                    DecimalIntegerValueModelElement("new_pruned"),
                    FixedDataModelElement("est_pruned_str", b',"est_pruned":'),
                    DecimalIntegerValueModelElement("est_pruned"),
                    FixedDataModelElement("bypassed_pruned_str", b',"bypassed_pruned":'),
                    DecimalIntegerValueModelElement("bypassed_pruned"),
                    FixedDataModelElement("flows_checked_str", b',"flows_checked":'),
                    DecimalIntegerValueModelElement("flows_checked"),
                    FixedDataModelElement("flows_notimeout_str", b',"flows_notimeout":'),
                    DecimalIntegerValueModelElement("flows_notimeout"),
                    FixedDataModelElement("flows_timeout_str", b',"flows_timeout":'),
                    DecimalIntegerValueModelElement("flows_timeout"),
                    FixedDataModelElement("flows_timeout_inuse_str", b',"flows_timeout_inuse":'),
                    DecimalIntegerValueModelElement("flows_timeout_inuse"),
                    FixedDataModelElement("flows_removed_str", b',"flows_removed":'),
                    DecimalIntegerValueModelElement("flows_removed"),
                    FixedDataModelElement("rows_checked_str", b',"rows_checked":'),
                    DecimalIntegerValueModelElement("rows_checked"),
                    FixedDataModelElement("rows_skipped_str", b',"rows_skipped":'),
                    DecimalIntegerValueModelElement("rows_skipped"),
                    FixedDataModelElement("rows_empty_str", b',"rows_empty":'),
                    DecimalIntegerValueModelElement("rows_empty"),
                    FixedDataModelElement("rows_busy_str", b',"rows_busy":'),
                    DecimalIntegerValueModelElement("rows_busy"),
                    FixedDataModelElement("rows_maxlen_str", b',"rows_maxlen":'),
                    DecimalIntegerValueModelElement("rows_maxlen"),
                    ]),
                SequenceModelElement("dns", [
                    FixedDataModelElement("memuse_str", b'},"dns":{"memuse":'),
                    DecimalIntegerValueModelElement("memuse"),
                    FixedDataModelElement("memcap_state_str", b',"memcap_state":'),
                    DecimalIntegerValueModelElement("memcap_state"),
                    FixedDataModelElement("memcap_global_str", b',"memcap_global":'),
                    DecimalIntegerValueModelElement("memcap_global"),
                    ]),
                SequenceModelElement("http", [
                    FixedDataModelElement("memuse_str", b'},"http":{"memuse":'),
                    DecimalIntegerValueModelElement("memuse"),
                    FixedDataModelElement("memcap_str", b',"memcap":'),
                    DecimalIntegerValueModelElement("memcap"),
                    ]),
                FixedDataModelElement("quote_str", b"}}}")
                ]),
            SequenceModelElement("tls", [
                FixedDataModelElement("tls_str", b'tls",'),
                conn,
                SequenceModelElement("tls", [
                    FixedDataModelElement("subject_str", b',"tls":{"subject":"'),
                    DelimitedDataModelElement("subject", b'"'),
                    FixedDataModelElement("issuerdn_str", b'","issuerdn":"'),
                    DelimitedDataModelElement("issuerdn", b'"'),
                    FixedDataModelElement("fingerprint_str", b'","fingerprint":"'),
                    DelimitedDataModelElement("fingerprint", b'"'),
                    OptionalMatchModelElement(
                        "sni", SequenceModelElement("sni", [
                            FixedDataModelElement("sni_str", b'","sni":"'),
                            DelimitedDataModelElement("sni", b'"'),
                            ])
                        ),
                    FixedDataModelElement("version_str", b'","version":"'),
                    DelimitedDataModelElement("version", b'"'),
                    FixedDataModelElement("notbefore_str", b'","notbefore":"'),
                    DelimitedDataModelElement("notbefore", b'"'),
                    FixedDataModelElement("notafter_str", b'","notafter":"'),
                    DelimitedDataModelElement("notafter", b'"'),
                    ]),
                FixedDataModelElement("brack_str", b'"}}')
                ]),
            SequenceModelElement("alert", [
                FixedDataModelElement("alert_str", b'alert",'),
                conn,
                OptionalMatchModelElement(
                    "tx_id", SequenceModelElement("tx_id", [
                        FixedDataModelElement("tx_id", b',"tx_id":'),
                        DecimalIntegerValueModelElement("tx_id"),
                        ])),
                SequenceModelElement("alert", [
                    FixedDataModelElement("action_str", b',"alert":{"action":"'),
                    DelimitedDataModelElement("action", b'"'),
                    FixedDataModelElement("gid_str", b'","gid":'),
                    DecimalIntegerValueModelElement("gid"),
                    FixedDataModelElement("signature_id_str", b',"signature_id":'),
                    DecimalIntegerValueModelElement("signature_id"),
                    FixedDataModelElement("rev_str", b',"rev":'),
                    DecimalIntegerValueModelElement("rev"),
                    FixedDataModelElement("signature_str", b',"signature":"'),
                    DelimitedDataModelElement("signature", b'"'),
                    FixedDataModelElement("category_str", b'","category":"'),
                    DelimitedDataModelElement("category", b'"'),
                    FixedDataModelElement("severity_str", b'","severity":'),
                    DecimalIntegerValueModelElement("severity"),
                    FixedDataModelElement("brack_str", b"}")
                    ]),
                http,
                FixedDataModelElement("brack_str", b"}")
                ]),
            ])
        ])

    return model



import unittest
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase, DummyMatchContext, DummyFixedDataModelElement


class OptionalMatchModelElementTest(TestBase):
    """Unittests for the OptionalMatchModelElement."""

    id_ = "optional"
    path = "path"
    fixed_id = "fixed"
    fixed_data = b"fixed data"

    def test1get_match_element_valid_match(self):
        """Parse matching substring from MatchContext and check if the MatchContext was updated with all characters."""
        data = b"fixed data string."
        value = self.fixed_data
        match_context = DummyMatchContext(data)
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        optional_match = OptionalMatchModelElement(self.id_, fixed_dme)
        match_element = optional_match.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, [
            fixed_dme.get_match_element("%s/%s" % (self.path, self.id_), DummyMatchContext(data))])

    def test2get_match_element_no_match(self):
        """Parse not matching substring from MatchContext and check if the MatchContext was not changed."""
        optional_match = OptionalMatchModelElement(self.id_, DummyFixedDataModelElement(self.fixed_id, self.fixed_data))
        data = b""
        match_context = DummyMatchContext(data)
        match_element = optional_match.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, None, None)

        data = b"other fixed string"
        value = b""
        match_context = DummyMatchContext(data)
        match_element = optional_match.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, None, None)

    def test3element_id_input_validation(self):
        """Check if element_id is validated."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        self.assertRaises(ValueError, OptionalMatchModelElement, "", fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, None, fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, b"path", fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, True, fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, 123, fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, 123.22, fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, {"id": "path"}, fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, ["path"], fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, [], fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, (), fixed_dme)
        self.assertRaises(TypeError, OptionalMatchModelElement, set(), fixed_dme)

    def test4optional_element_input_validation(self):
        """Check if optional_element is validated."""
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, "fixed_dme")
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, None)
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, b"path")
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, True)
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, 123)
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, 123.22)
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, {"id": "path"})
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, ["path"])
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, [])
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, ())
        self.assertRaises(TypeError, OptionalMatchModelElement, self.id_, set())

    def test5get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = OptionalMatchModelElement(self.id_, DummyFixedDataModelElement(self.fixed_id, self.fixed_data))
        data = b"fixed data"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""This module defines a parser model for exim."""

from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement


def get_model():
    """Return a model to parse Exim logs from the AIT-LDS."""
    alphabet = b"!'#$%&\"()*+,-./0123456789:;<>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ\\^_`abcdefghijklmnopqrstuvwxyz{|}~=[]"

    model = SequenceModelElement("model", [
        DateTimeModelElement("time", b"%Y-%m-%d %H:%M:%S"),
        FixedDataModelElement("sp", b" "),
        FirstMatchModelElement("fm", [
            SequenceModelElement("start", [
                FixedDataModelElement("start", b"Start queue run: pid="),
                DecimalIntegerValueModelElement("pid"),
                ]),
            SequenceModelElement("start", [
                FixedDataModelElement("start", b"TLS error"),
                AnyByteDataModelElement('remainder')
                ]),
            SequenceModelElement("end", [
                FixedDataModelElement("end", b"End queue run: pid="),
                DecimalIntegerValueModelElement("pid"),
                ]),
            SequenceModelElement("no_host_found", [
                FixedDataModelElement("no_host_found_str", b"no host name found for IP address "),
                IpAddressDataModelElement("ip"),
                ]),
            SequenceModelElement("vrfy_failed", [
                FixedDataModelElement("vrfy_failed_str", b"VRFY failed for "),
                DelimitedDataModelElement("mail", b" "),
                FixedDataModelElement("h_str", b" H="),
                DelimitedDataModelElement("h", b" "),
                FixedDataModelElement("sp1", b" ["),
                IpAddressDataModelElement("ip"),
                FixedDataModelElement("sp2", b"]")
                ]),
            SequenceModelElement("mail", [
                DelimitedDataModelElement("id", b" "),
                FirstMatchModelElement("dir", [
                    SequenceModelElement("dir_in", [
                        FixedDataModelElement("in", b" <= "),
                        FirstMatchModelElement("fm", [
                            SequenceModelElement("seq1", [
                                FixedDataModelElement("brack", b"<> "),
                                FirstMatchModelElement("fm", [
                                    SequenceModelElement("r", [
                                        FixedDataModelElement("r_str", b"R="),
                                        DelimitedDataModelElement("r", b" "),
                                        FixedDataModelElement("u_str", b" U="),
                                        DelimitedDataModelElement("u", b" "),
                                        ]),
                                    SequenceModelElement("h", [
                                        FixedDataModelElement("h_str", b"H="),
                                        DelimitedDataModelElement("h", b" "),
                                        FixedDataModelElement("sp1", b" ["),
                                        IpAddressDataModelElement("ip"),
                                        FixedDataModelElement("sp1", b"]"),
                                        ])
                                    ]),
                                FixedDataModelElement("sp2", b" P="),
                                DelimitedDataModelElement("p", b" "),
                                FixedDataModelElement("sp2", b" S="),
                                DecimalIntegerValueModelElement("s"),
                                ]),
                            SequenceModelElement("seq2", [
                                DelimitedDataModelElement("mail", b" "),
                                FixedDataModelElement("user_str", b" U="),
                                DelimitedDataModelElement("user", b" "),
                                FixedDataModelElement("p_str", b" P="),
                                DelimitedDataModelElement("p", b" "),
                                FixedDataModelElement("s_str", b" S="),
                                DecimalIntegerValueModelElement("s"),
                                OptionalMatchModelElement(
                                    "id", SequenceModelElement("id", [
                                        FixedDataModelElement("id_str", b" id="),
                                        AnyByteDataModelElement("id")
                                        ])
                                    )
                                ]),
                            AnyByteDataModelElement('remainder')
                            ])
                        ]),
                    SequenceModelElement("dir_out", [
                        FixedDataModelElement("in", b" => "),
                        DelimitedDataModelElement("name", b" "),
                        FirstMatchModelElement('fm', [
                            SequenceModelElement('seq', [
                                FixedDataModelElement("sp1", b" "),
                                OptionalMatchModelElement(
                                    "mail_opt",
                                    SequenceModelElement(
                                        "mail", [
                                            FixedDataModelElement("brack1", b"("),
                                            DelimitedDataModelElement("brack_mail", b")"),
                                            FixedDataModelElement("brack2", b") "),
                                        ])
                                ),
                                OptionalMatchModelElement(
                                    "opt",
                                    SequenceModelElement(
                                        "seq", [
                                            FixedDataModelElement("sp2", b"<"),
                                            DelimitedDataModelElement("mail", b">"),
                                            FixedDataModelElement("closing_brack", b"> "),
                                        ])
                                ),
                                FixedDataModelElement("r_str", b"R="),
                                DelimitedDataModelElement("r", b" "),
                                FixedDataModelElement("t_str", b" T="),
                                VariableByteDataModelElement("t", alphabet),
                                OptionalMatchModelElement(
                                    "param_opt",
                                    SequenceModelElement(
                                        "seq", [
                                            FixedDataModelElement("h_str", b" H="),
                                            DelimitedDataModelElement("h", b" X="),
                                            FixedDataModelElement("x_str", b" X="),
                                            DelimitedDataModelElement("x", b" CV="),
                                            FixedDataModelElement("cv_str", b" CV="),
                                            DelimitedDataModelElement("cv", b" DN="),
                                            FixedDataModelElement("dn_str", b" DN="),
                                            DelimitedDataModelElement("dn", b" C="),
                                            AnyByteDataModelElement("c"),
                                        ]))
                                ]),
                            ])
                        ]),
                    SequenceModelElement("aster", [
                        FixedDataModelElement("aster", b" ** "),
                        DelimitedDataModelElement("command", b" "),
                        FixedDataModelElement("headers_str", b' Too many "Received" headers - suspected mail loop')]),
                    SequenceModelElement("prdr", [
                        FixedDataModelElement("prdr", b" PRDR "),
                        AnyByteDataModelElement('remainder')]),
                    SequenceModelElement("arrw", [
                        FixedDataModelElement("arrw", b" -> "),
                        AnyByteDataModelElement('remainder')]),
                    FixedDataModelElement("completed", b" Completed"),
                    FixedDataModelElement("frozen", b" Message is frozen"),
                    FixedDataModelElement("frozen", b" Frozen (delivery error message)")
                ])
                ])])])

    return model



"""
This module defines a detector for new data paths.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface
from aminer.util import PersistenceUtil


class NewMatchPathDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class creates events when new data path was found in a parsed atom."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, anomaly_event_handlers, persistence_id='Default', learn_mode=False, output_logline=True,
                 stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param persistence_id name of persistence file.
        @param learn_mode specifies whether new values should be learned.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            aminer_config=aminer_config, anomaly_event_handlers=anomaly_event_handlers, persistence_id=persistence_id,
            learn_mode=learn_mode, output_logline=output_logline,  stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.log_learned_paths = 0
        self.log_new_learned_paths = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is None:
            self.known_path_set = set()
        else:
            self.known_path_set = set(persistence_data)
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    def receive_atom(self, log_atom):
        """
        Receive on parsed atom and the information about the parser match.
        @param log_atom the parsed log atom
        @return True if this handler was really able to handle and process the match. Depending on this information, the caller
                may decide if it makes sense passing the parsed atom also to other handlers.
        """
        self.log_total += 1
        unknown_path_list = []
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        for path in log_atom.parser_match.get_match_dictionary().keys():
            if path not in self.known_path_set:
                unknown_path_list.append(path)
                if self.learn_mode:
                    self.known_path_set.add(path)
                    self.log_learned_paths += 1
                    self.log_new_learned_paths.append(path)
                    if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                        self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time
        if unknown_path_list:
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            if self.output_logline:
                sorted_log_lines = [log_atom.parser_match.match_element.annotate_match('') + os.linesep + repr(
                    unknown_path_list) + os.linesep + original_log_line_prefix + data]
            else:
                sorted_log_lines = [repr(unknown_path_list) + os.linesep + original_log_line_prefix + data]
            analysis_component = {'AffectedLogAtomPaths': list(unknown_path_list)}
            event_data = {'AnalysisComponent': analysis_component}
            for listener in self.anomaly_event_handlers:
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New path(es) detected', sorted_log_lines, event_data,
                                       log_atom, self)
        self.log_success += 1
        return True

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        PersistenceUtil.store_json(self.persistence_file_name, list(self.known_path_set))
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        self.known_path_set.add(event_data)
        return f'Allowlisted path(es) {event_data} in {event_type}.'

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_paths} new paths in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_paths} new paths in the last 60 minutes. Following new paths were learned: "
                f"{self.log_new_learned_paths}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned_paths = 0
        self.log_new_learned_paths = []



import unittest
from aminer.analysis.EventCountClusterDetector import EventCountClusterDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomalies = []

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomalies.append(evdat)


class EventCountClusterDetectorTest(TestBase):
    """Unittests for the EventFrequencyDetector."""

    def test1_normal_count_detection(self):
        """
        This test checks the normal operation of the ECCD
        """
        description = "Test1EventFrequencyDetector"

        # Initialize detector for analyzing values in one path in time windows of 10 seconds
        test_handler = TestHandler()
        eccd = EventCountClusterDetector(aminer_config=self.aminer_config, anomaly_event_handlers=[test_handler],
                                                          target_path_list=['/p/value'], id_path_list=['/p/id'],
                                                          window_size=10, num_windows=50,
                                                          confidence_factor=0.5, idf=True, norm=False, add_normal=False,
                                                          check_empty_windows=False, persistence_id='Default',
                                                          learn_mode=True, output_logline=False)
        self.analysis_context.register_component(eccd, description)

        # The following log atoms are created:
        #  window 1:
        #   value a: 1 time by x, 1 time by y
        #   value b: 1 time by x
        #  window 2:
        #   value a: 2 times by x, 1 time by y
        #   value b: 1 time by x
        #  window 3:
        #   value b: 1 time by x
        #   value c: 1 time by x
        #  window 4:
        #   value a: 1 time by x
        # Start of window 1:
        m1_1 = MatchElement('/p/value', b'a', b'a', None)
        m1_2 = MatchElement('/p/id', b'x', b'x', None)
        p1 = ParserMatch(MatchElement('/p', b'ax', b'ax', [m1_1, m1_2]))
        log_atom_1 = LogAtom(b'ax', p1, 1, None)

        m2_1 = MatchElement('/p/value', b'a', b'a', None)
        m2_2 = MatchElement('/p/id', b'y', b'y', None)
        p2 = ParserMatch(MatchElement('/p', b'ay', b'ay', [m2_1, m2_2]))
        log_atom_2 = LogAtom(b'ay', p2, 2, None)

        m3_1 = MatchElement('/p/value', b'b', b'b', None)
        m3_2 = MatchElement('/p/id', b'x', b'x', None)
        p3 = ParserMatch(MatchElement('/p', b'bx', b'bx', [m3_1, m3_2]))
        log_atom_3 = LogAtom(b'bx', p3, 3, None)

        # Start of window 2:
        m4_1 = MatchElement('/p/value', b'a', b'a', None)
        m4_2 = MatchElement('/p/id', b'x', b'x', None)
        p4 = ParserMatch(MatchElement('/p', b'ax', b'ax', [m4_1, m4_2]))
        log_atom_4 = LogAtom(b'ax', p4, 13, None)

        m5_1 = MatchElement('/p/value', b'a', b'a', None)
        m5_2 = MatchElement('/p/id', b'y', b'y', None)
        p5 = ParserMatch(MatchElement('/p', b'ay', b'ay', [m5_1, m5_2]))
        log_atom_5 = LogAtom(b'ay', p5, 14, None)

        m6_1 = MatchElement('/p/value', b'b', b'b', None)
        m6_2 = MatchElement('/p/id', b'x', b'x', None)
        p6 = ParserMatch(MatchElement('/p', b'bx', b'bx', [m6_1, m6_2]))
        log_atom_6 = LogAtom(b'bx', p6, 15, None)

        m7_1 = MatchElement('/p/value', b'a', b'a', None)
        m7_2 = MatchElement('/p/id', b'x', b'x', None)
        p7 = ParserMatch(MatchElement('/p', b'ax', b'ax', [m7_1, m7_2]))
        log_atom_7 = LogAtom(b'ax', p4, 16, None)

        # Start of window 3:
        m8_1 = MatchElement('/p/value', b'c', b'c', None)
        m8_2 = MatchElement('/p/id', b'x', b'x', None)
        p8 = ParserMatch(MatchElement('/p', b'cx', b'cx', [m8_1, m8_2]))
        log_atom_8 = LogAtom(b'cx', p8, 23, None)

        m9_1 = MatchElement('/p/value', b'b', b'b', None)
        m9_2 = MatchElement('/p/id', b'x', b'x', None)
        p9 = ParserMatch(MatchElement('/p', b'bx', b'bx', [m9_1, m9_2]))
        log_atom_9 = LogAtom(b'bx', p9, 24, None)

        # Start of window 4:
        m10_1 = MatchElement('/p/value', b'a', b'a', None)
        m10_2 = MatchElement('/p/id', b'x', b'x', None)
        p10 = ParserMatch(MatchElement('/p', b'ax', b'ax', [m10_1, m10_2]))
        log_atom_10 = LogAtom(b'ax', p10, 43, None)

        # Forward log atoms to detector
        eccd.receive_atom(log_atom_1)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_2)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_3)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_4)
        # End of first time window; first count vector triggers anomaly for x
        self.assertTrue(test_handler.anomalies)
        # Remove anomaly
        test_handler.anomalies = []
        eccd.receive_atom(log_atom_5)
        # End of first time window; first count vector triggers anomaly for y
        self.assertTrue(test_handler.anomalies)
        # Remove anomaly
        test_handler.anomalies = []
        eccd.receive_atom(log_atom_6)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_7)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_8)
        # No anomaly reported for x since 2 times a and 1 time b (window 1) is similar enough to 1 time a and 1 time b (window 2)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_9)
        self.assertFalse(test_handler.anomalies)
        eccd.receive_atom(log_atom_10)
        # Check learned count vectors at end of third time window
        # For x, count vector from first and third windows are included in model; for y only first window
        self.assertEquals(eccd.known_counts, {('x',): [{('a',): 1, ('b',): 1}, {('c',): 1, ('b',): 1}], ('y',): [{('a',): 1}]})
        # Since a occurs in both x and y, its idf factor is only 0.176 (=log10(3/2)),
        # compared to b and c which have an idf factor of 0.477 (=log10(3/1)).
        # Comparing the count vectors for x in the first and third window, we see that
        #  a occurs only in first window, which increases diff to 0.176/0.176
        #  b occurs once in first and third windows, which updates diff to 0.176/0.653
        #  c occurs only in third window, which increaes diff to 0.653/1.13
        # The final score is thus 0.653/1.13=0.578, which exceeds the threshold of 0.5.
        self.assertEqual(test_handler.anomalies, [
            {'AnalysisComponent':
                {'AffectedLogAtomPaths': ['/p/value'],
                 'AffectedLogAtomFrequencies': [1, 1],
                 'AffectedIdValues': ['x'],
                 'AffectedLogAtomValues': [('c',), ('b',)]}, 'CountData': {
                     'ConfidenceFactor': 0.5, 'Confidence': 0.577893478883737
                     }}])

if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.analysis.CharsetDetector import CharsetDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomaly = None

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomaly = evdat


class CharsetDetectorTest(TestBase):
    """Unittests for the ValueRangeDetectorDetector."""

    def test1_normal_sequence_detection(self):
        """
        This test case checks the normal detection of new character sets.
        The charset detector is used to learn an alphabet and detect new characters for different identifiers.
        """
        description = "Test1CharsetDetector"

        # Initialize detector
        test_handler = TestHandler()
        event_charset_detector = CharsetDetector(self.aminer_config, [test_handler], ['/model/id'], ['/model/value'], 'Default', True,
                                                 False)
        self.analysis_context.register_component(event_charset_detector, description)

        # Prepare log atoms that represent two entities (id) with strings (value). Anomalies are generated when new characters are observed.
        # The following events are generated:
        #  id: a value: abc
        #  id: b value: xyz
        #  id: a value: asdf
        #  id: a value: xxx
        #  id: a value: bass
        #  id: a value: max
        m_1 = MatchElement('/model/id', b'a', b'a', None)
        m_2 = MatchElement('/model/value', b'abc', b'abc', None)
        match_element_1 = MatchElement('/model', b'aabc', b'aabc', [m_1, m_2])
        parser_match_1 = ParserMatch(match_element_1)
        log_atom_1 = LogAtom(b'aabc', parser_match_1, 1, None)

        m_3 = MatchElement('/model/id', b'b', b'b', None)
        m_4 = MatchElement('/model/value', b'xyz', b'xyz', None)
        match_element_2 = MatchElement('/model', b'bxyz', b'bxyz', [m_3, m_4])
        parser_match_2 = ParserMatch(match_element_2)
        log_atom_2 = LogAtom(b'bxyz', parser_match_2, 2, None)

        m_5 = MatchElement('/model/id', b'a', b'a', None)
        m_6 = MatchElement('/model/value', b'asdf', b'asdf', None)
        match_element_3 = MatchElement('/model', b'aasdf', b'aasdf', [m_5, m_6])
        parser_match_3 = ParserMatch(match_element_3)
        log_atom_3 = LogAtom(b'aasdf', parser_match_3, 3, None)

        m_7 = MatchElement('/model/id', b'a', b'a', None)
        m_8 = MatchElement('/model/value', b'xxx', b'xxx', None)
        match_element_4 = MatchElement('/model', b'bxxx', b'bxxx', [m_7, m_8])
        parser_match_4 = ParserMatch(match_element_4)
        log_atom_4 = LogAtom(b'bxxx', parser_match_4, 4, None)

        m_9 = MatchElement('/model/id', b'a', b'a', None)
        m_10 = MatchElement('/model/value', b'bass', b'bass', None)
        match_element_5 = MatchElement('/model', b'abass', b'abass', [m_9, m_10])
        parser_match_5 = ParserMatch(match_element_5)
        log_atom_5 = LogAtom(b'abass', parser_match_5, 5, None)

        m_11 = MatchElement('/model/id', b'a', b'a', None)
        m_12 = MatchElement('/model/value', b'max', b'max', None)
        match_element_6 = MatchElement('/model', b'bmax', b'bmax', [m_11, m_12])
        parser_match_6 = ParserMatch(match_element_6)
        log_atom_6 = LogAtom(b'bmax', parser_match_6, 6, None)

        # Forward log atoms to detector
        # First value of id (a) should not generate an anomaly
        # Input: id: a value: abc
        # Expected output: None
        event_charset_detector.receive_atom(log_atom_1)
        self.assertIsNone(test_handler.anomaly)

        # First value of id (b) should not generate an anomaly
        # Input: id: b value: xyz
        # Expected output: None
        event_charset_detector.receive_atom(log_atom_2)
        self.assertIsNone(test_handler.anomaly)

        # Second value of id (a) should generate an anomaly for new characters ('sdf' of 'asdf' not in 'abc')
        # Input: id: a value: asdf
        # Expected output: Anomaly
        event_charset_detector.receive_atom(log_atom_3)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/model/value'],
                                                                      'AffectedLogAtomValues': ['asdf'],
                                                                      'MissingCharacters': ['s', 'd', 'f']}})
        test_handler.anomaly = None

        # Third value of id (a) should generate an anomaly for new characters ('x' not in 'abcsdf', only in 'xyz' from other id (b))
        # Input: id: a value: xxx
        # Expected output: Anomaly
        event_charset_detector.receive_atom(log_atom_4)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/model/value'],
                                                                      'AffectedLogAtomValues': ['xxx'],
                                                                      'MissingCharacters': ['x']}})
        test_handler.anomaly = None

        # Fourth value of id (a) should not generate an anomaly (all characters of 'bass' in 'abcsdfx')
        # Input: id: a value: bass
        # Expected output: None
        event_charset_detector.receive_atom(log_atom_5)
        self.assertIsNone(test_handler.anomaly)

        # Fifth value of id (a) should generate an anomaly for new characters ('m' of 'max' not in 'abcsdfx')
        event_charset_detector.receive_atom(log_atom_6)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/model/value'],
                                                                      'AffectedLogAtomValues': ['max'],
                                                                      'MissingCharacters': ['m']}})
        test_handler.anomaly = None


if __name__ == "__main__":
    unittest.main()



{
    'EventHandlers': {
        'required': False,
        'type': 'list',
        'nullable': True,
        'default': None,
        'schema': {
            'type': 'dict',
            'schema': {
                'id': {'type': 'string', 'required': True},
                'type': {'type': 'eventhandlertype', 'coerce': 'toeventhandlertype', 'required': True},
                'json': {'type': 'boolean', 'default': False},
                'score': {'type': 'boolean', 'default': False},
                'instance_name': {'type': 'string', 'default': 'aminer'},
                'topic': {'type': 'string'},
                'url': {'type': 'string', 'default': 'ipc:///tmp/aminer'},
                'cfgfile': {'type': 'string', 'default': '/etc/aminer/kafka-client.conf'},
                'options': {'type': 'dict', 'schema': {'id': {'type': 'string'}, 'type': {'type': ['string', 'list', 'integer']}}},
                'output_file_path': {'type': 'string'},
                'pretty': {'type': 'boolean', 'default': True},
                'weights': {'type': 'dict', 'nullable': True, 'default': None},
                'auto_weights': {'type': 'boolean', 'default': False},
                'auto_weights_history_length': {'type': 'integer', 'default': 1000, 'min': 1}
            }
        }
    }
}



{
    'EventHandlers': {
        'required': False,
        'type': 'list',
        'nullable': True,
        'default': None,
        'schema': {
            'type': 'dict',
            'allow_unknown': False,
            'oneof_schema': [
                {
                    'id': {'type': 'string', 'required': True, 'empty': False},
                    'type': {'type': 'string', 'forbidden': [
                        'KafkaEventHandler', 'ZmqEventHandler', 'StreamPrinterEventHandler', 'SyslogWriterEventHandler'], 'required': True},
                    'json': {'type': 'boolean'},
                    'score': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'required': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ZmqEventHandler'], 'required': True},
                    'json': {'type': 'boolean'},
                    'pretty': {'type': 'boolean'},
                    'score': {'type': 'boolean'},
                    'weights': {'type': 'dict', 'nullable': True},
                    'auto_weights': {'type': 'boolean'},
                    'auto_weights_history_length': {'type': 'integer', 'default': 1000, 'min': 1},
                    'topic': {'type': 'string', 'required': False},
                    'url': {'type': 'string', 'empty': False},
                },
                {
                    'id': {'type': 'string', 'required': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['KafkaEventHandler'], 'required': True},
                    'json': {'type': 'boolean'},
                    'pretty': {'type': 'boolean'},
                    'score': {'type': 'boolean'},
                    'weights': {'type': 'dict', 'nullable': True},
                    'auto_weights': {'type': 'boolean'},
                    'auto_weights_history_length': {'type': 'integer', 'default': 1000, 'min': 1},
                    'topic': {'type': 'string', 'required': True, 'empty': False},
                    'cfgfile': {'type': 'string', 'empty': False},
                    'options': {'type': 'dict', 'schema': {
                        'id': {'type': 'string', 'empty': False}, 'type': {'type': ['string', 'list', 'integer']}}},
                },
                {
                    'id': {'type': 'string', 'required': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['StreamPrinterEventHandler'], 'required': True},
                    'json': {'type': 'boolean'},
                    'pretty': {'type': 'boolean'},
                    'score': {'type': 'boolean'},
                    'weights': {'type': 'dict', 'nullable': True},
                    'auto_weights': {'type': 'boolean'},
                    'auto_weights_history_length': {'type': 'integer', 'default': 1000, 'min': 1},
                    'output_file_path': {'type': 'string', 'empty': False}
                },
                {
                    'id': {'type': 'string', 'required': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['SyslogWriterEventHandler'], 'required': True},
                    'json': {'type': 'boolean'},
                    'pretty': {'type': 'boolean'},
                    'score': {'type': 'boolean'},
                    'weights': {'type': 'dict', 'nullable': True},
                    'auto_weights': {'type': 'boolean'},
                    'auto_weights_history_length': {'type': 'integer', 'default': 1000, 'min': 1},
                    'instance_name': {'type': 'string', 'default': 'aminer', 'empty': False}
                }
            ]
        }
    }
}



import unittest
from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
from time import time
from aminer.parsing.ParserMatch import ParserMatch
from aminer.parsing.MatchContext import MatchContext
from aminer.input.LogAtom import LogAtom
from aminer.analysis.UnparsedAtomHandlers import SimpleUnparsedAtomHandler
from unit.TestBase import TestBase


class SimpleUnparsedAtomHandlerTest(TestBase):
    """Unittests for the SimpleUnparsedAtomHandler."""

    calculation = b'256 * 2 = 512'

    def test1_atom_is_unparsed(self):
        """The atom in this test case has a ParserMatch."""
        description = "Test1SimpleUnparsedAtomHandler"
        any_byte_data_model_element = AnyByteDataModelElement('a1')
        new_match_path_detector1 = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)

        match_context = MatchContext(self.calculation)
        match_element = any_byte_data_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_element.match_object, ParserMatch(match_element), time(), new_match_path_detector1)

        simple_unparsed_atom_handler = SimpleUnparsedAtomHandler([self.stream_printer_event_handler])
        self.analysis_context.register_component(simple_unparsed_atom_handler, description)
        self.assertTrue(not simple_unparsed_atom_handler.receive_atom(log_atom))

    def test2_atom_is_parsed(self):
        """The atom in this test case has no ParserMatch."""
        description = "Test2SimpleUnparsedAtomHandler"
        any_byte_data_model_element = AnyByteDataModelElement('a1')
        new_match_path_detector1 = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)

        match_context = MatchContext(self.calculation)
        match_element = any_byte_data_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_element.match_object, None, time(), new_match_path_detector1)

        simple_unparsed_atom_handler = SimpleUnparsedAtomHandler([self.stream_printer_event_handler])
        self.analysis_context.register_component(simple_unparsed_atom_handler, description)
        self.assertTrue(simple_unparsed_atom_handler.receive_atom(log_atom))

    def test3_parser_match_is_other_element(self):
        """In this test case the ParserMatch actually is no instance of ParserMatch. The atom should still be considered to be parsed."""
        description = "Test3SimpleUnparsedAtomHandler"
        any_byte_data_model_element = AnyByteDataModelElement('a1')
        new_match_path_detector1 = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)

        match_context = MatchContext(self.calculation)
        match_element = any_byte_data_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_element.match_object, any_byte_data_model_element, time(), new_match_path_detector1)

        simple_unparsed_atom_handler = SimpleUnparsedAtomHandler([self.stream_printer_event_handler])
        self.analysis_context.register_component(simple_unparsed_atom_handler, description)
        self.assertTrue(not simple_unparsed_atom_handler.receive_atom(log_atom))


if __name__ == "__main__":
    unittest.main()



"""This module contains multiple History classes used by the aminer.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.

"""
import random
import abc
from aminer.input.InputInterfaces import AtomHandlerInterface


def get_log_int(max_bits):
    """Get a log-distributed random integer integer in range 0 to maxBits-1."""
    rand_bits = random.randint(0, (1 << max_bits) - 1)
    result = 0
    while (rand_bits & 1) != 0:
        result += 1
        rand_bits >>= 1
    return result


class ObjectHistory(metaclass=abc.ABCMeta):
    """
    This is the superinterface of all object histories.
    The idea behind that is to use that type of history best suited for a purpose considering amount of data, possibility for history size
    limits to be reached, priorization which elements should be dropped first.
    """

    @abc.abstractmethod
    def add_object(self, new_object):
        """Add an object to this history. This method call may evict other objects from the history."""

    @abc.abstractmethod
    def get_history(self):
        """Get the whole history list. Make sure to clone the list before modification when influences on this object are not intended."""

    @abc.abstractmethod
    def clear_history(self):
        """Clean the whole history."""


class LogarithmicBackoffHistory(ObjectHistory):
    """
    This class keeps a history list of items with logarithmic storage characteristics.
    When adding objects, the list will be filled to the maximum size with the newest items at the end. When filled, adding a new element
    will replace with probability 1/2 the last element. With a chance of 1/4, the last element will be moved to the next lower position,
    before putting the new element at the end of the list. With a chance of 1/8, the last two elements are moved, ... Thus the list will in
    average span a time range of 2^maxItems items with growing size of holes towards the earliest element.
    """

    def __init__(self, max_items, initial_list=None):
        self.max_items = max_items
        if initial_list is None:
            initial_list = []
        else:
            initial_list = initial_list[:max_items]
        self.history = initial_list

    def add_object(self, new_object):
        """Add a new object to the list according to the rules described in the class docstring."""
        if len(self.history) < self.max_items:
            self.history.append(new_object)
        else:
            move_pos = get_log_int(self.max_items - 1)
            self.history = self.history[:self.max_items - move_pos - 1] + self.history[self.max_items - move_pos:] + [new_object]

    def get_history(self):
        """Get the whole history list. Make sure to clone the list before modification when influences on this object are not intended."""
        return self.history

    def clear_history(self):
        """Clean the whole history."""
        self.history[:] = []


class VolatileLogarithmicBackoffAtomHistory(AtomHandlerInterface, LogarithmicBackoffHistory):
    """
    This class is a volatile filter to keep a history of log atoms.
    Example usages can be for analysis by other components or for external access via remote control interface.
    """

    def __init__(self, max_items):
        """Initialize the history component."""
        LogarithmicBackoffHistory.__init__(self, max_items)

    def receive_atom(self, log_atom):
        """Receive an atom and add it to the history log."""
        self.add_object(log_atom)
        return True



"""This module defines a parser for ssmtp."""

from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement


def get_model():
    """Return the model."""
    type_children = [
        SequenceModelElement("sent", [
            FixedDataModelElement("s0", b"Sent mail for "),
            DelimitedDataModelElement("to-addr", b" ("),
            FixedDataModelElement("s1", b" ("),
            DelimitedDataModelElement("status", b") uid="),
            FixedDataModelElement("s2", b") uid="),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s3", b" username="),
            DelimitedDataModelElement("username", b" outbytes="),
            FixedDataModelElement("s4", b" outbytes="),
            DecimalIntegerValueModelElement("bytes")
        ]),
        SequenceModelElement("sent", [
            DelimitedDataModelElement("program", b" "),
            FixedDataModelElement("s0", b" sent mail for "),
            AnyByteDataModelElement("user")
        ])
    ]

    model = SequenceModelElement("ssmtp", [
        FixedDataModelElement("sname", b"sSMTP["),
        DecimalIntegerValueModelElement("pid"),
        FixedDataModelElement("s0", b"]: "),
        FirstMatchModelElement("msg", type_children)
    ])
    return model



import unittest
from aminer.analysis.EventFrequencyDetector import EventFrequencyDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomalies = []

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomalies.append(evdat)


class EventFrequencyDetectorTest(TestBase):
    """Unittests for the EventFrequencyDetector."""

    def test1_normal_frequency_detection(self):
        """
        This test case checks the normal detection of new frequencies.
        The EFD is used with one path to be analyzed over four time windows. The frequencies
        do not change a lot in the first time windows, thus no anomalies are generated. Then,
        value frequencies change and anomalies are created in the last time windows.
        """
        description = "Test1EventFrequencyDetector"

        # Initialize detector for analyzing values in one path in time windows of 10 seconds
        test_handler = TestHandler()
        event_frequency_detector = EventFrequencyDetector(aminer_config=self.aminer_config, anomaly_event_handlers=[test_handler],
                                                          target_path_list=['/value'], window_size=10, num_windows=1,
                                                          confidence_factor=0.51, empty_window_warnings=True, persistence_id='Default',
                                                          learn_mode=True, output_logline=False)
        self.analysis_context.register_component(event_frequency_detector, description)

        # Prepare log atoms that represent different amounts of values a, b over time
        # Four time windows are used. The first time window is used for initialization. The
        # second time window represents normal behavior, i.e., the frequencies do not change
        # too much and no anomalies should be generated. The third window contains changes
        # of value frequencies and thus anomalies should be generated. The fourth time window
        # only has the purpose of marking the end of the third time window.
        # The following log atoms are created:
        #  window 1:
        #   value a: 2 times
        #   value b: 1 time
        #  window 2:
        #   value a: 3 times
        #   value b: 1 time
        #  window 3:
        #   value a: 0 times
        #   value b: 2 times
        #  window 4:
        #   value a: 1 time
        # Start of window 1:
        m_1 = MatchElement('/value', b'a', b'a', None)
        parser_match_1 = ParserMatch(m_1)
        log_atom_1 = LogAtom(b'a', parser_match_1, 1, None)

        m_2 = MatchElement('/value', b'b', b'b', None)
        parser_match_2 = ParserMatch(m_2)
        log_atom_2 = LogAtom(b'b', parser_match_2, 3, None)

        m_3 = MatchElement('/value', b'a', b'a', None)
        parser_match_3 = ParserMatch(m_3)
        log_atom_3 = LogAtom(b'a', parser_match_3, 7, None)

        # Start of window 2:
        m_4 = MatchElement('/value', b'a', b'a', None)
        parser_match_4 = ParserMatch(m_4)
        log_atom_4 = LogAtom(b'a', parser_match_4, 13, None)

        m_6 = MatchElement('/value', b'b', b'b', None)
        parser_match_6 = ParserMatch(m_6)
        log_atom_6 = LogAtom(b'b', parser_match_6, 17, None)

        m_7 = MatchElement('/value', b'a', b'a', None)
        parser_match_7 = ParserMatch(m_7)
        log_atom_7 = LogAtom(b'a', parser_match_7, 18, None)

        m_8 = MatchElement('/value', b'a', b'a', None)
        parser_match_8 = ParserMatch(m_8)
        log_atom_8 = LogAtom(b'a', parser_match_8, 19, None)

        # Start of window 3:
        m_9 = MatchElement('/value', b'b', b'b', None)
        parser_match_9 = ParserMatch(m_9)
        log_atom_9 = LogAtom(b'b', parser_match_9, 25, None)

        m_10 = MatchElement('/value', b'b', b'b', None)
        parser_match_10 = ParserMatch(m_10)
        log_atom_10 = LogAtom(b'b', parser_match_10, 25, None)

        # Start of window 4:
        m_11 = MatchElement('/value', b'a', b'a', None)
        parser_match_11 = ParserMatch(m_11)
        log_atom_11 = LogAtom(b'a', parser_match_11, 35, None)

        # Forward log atoms to detector
        # Log atoms of initial window 1 should not create anomalies and add to counts
        # Input: a; initial time window is started
        # Expected output: frequency of a is 1
        event_frequency_detector.receive_atom(log_atom_1)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [1]})

        # Input: b; initial time window is not finished
        # Expected output: frequency of b is 1 added to existing count
        event_frequency_detector.receive_atom(log_atom_2)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [1], ('b',): [1]})

        # Input: a; initial time window is not finished
        # Expected output: frequency of a is 2 replaces a in existing count
        event_frequency_detector.receive_atom(log_atom_3)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2], ('b',): [1]})

        # Time window 2 should not create anomalies since a is in confidence (3 vs 2 occurrences) and b is identical (1 occurrence).
        # Input: a; initial time window is completed, second time window is started
        # Expected output: frequency of a is 1 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_4)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 1], ('b',): [1, 0]})

        # Input: b; second time window is not finished
        # Expected output: frequency of b is 1 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_6)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 1], ('b',): [1, 1]})

        # Input: a; second time window is not finished
        # Expected output: frequency of a is 3 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_7)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 2], ('b',): [1, 1]})

        # Input: a; second time window is not finished
        # Expected output: frequency of a is 4 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_8)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 3], ('b',): [1, 1]})

        # Time window 3 should create 2 anomalies since a drops from 3 to 0 and b increases from 1 to 2, which will be reported in window 4.
        # Anomalies are only reported when third time window is known to be completed, which will occur when subsequent atom is received.
        # Input: b; second time window is completed, third time window is started
        # Expected output: frequency of b is 1 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_9)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 3, 0], ('b',): [1, 1, 1]})

        # Input: b; third ime window is not finished
        # Expected output: frequency of b is 2 in new time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_10)
        self.assertFalse(test_handler.anomalies)
        self.assertEqual(event_frequency_detector.counts, {('a',): [2, 3, 0], ('b',): [1, 1, 2]})

        # Time window 4 should not create anomalies since no log atom is received to evaluate it.
        # Input: a; third time window is completed, fourth time window is started
        # Expected output: Anomalies for unexpected low counts of a (0 instead of 3) and b (2 instead of 1), frequency of a is 1 in new
        # time window count, old count remains unchanged
        event_frequency_detector.receive_atom(log_atom_11)
        self.assertEqual(test_handler.anomalies, [
            {'AnalysisComponent':
                {'AffectedLogAtomPaths': ['/value'],
                 'AffectedLogAtomValues': ['a']}, 'FrequencyData': {
                     'ExpectedLogAtomValuesFrequency': 3.0,
                     'ExpectedLogAtomValuesFrequencyRange': [1.0, 5.0],
                     'LogAtomValuesFrequency': 0,
                     'WindowSize': 10, 'ConfidenceFactor': 0.51, 'Confidence': 1.0
                     }}, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/value'],
                          'AffectedLogAtomValues': ['b']}, 'FrequencyData':
                              {'ExpectedLogAtomValuesFrequency': 1.0,
                               'ExpectedLogAtomValuesFrequencyRange': [1.0, 1.0],
                               'LogAtomValuesFrequency': 2,
                               'WindowSize': 10,
                               'ConfidenceFactor': 0.51,
                               'Confidence': 0.5}}])
        self.assertEqual(event_frequency_detector.counts, {('a',): [3, 0, 1], ('b',): [1, 2, 0]})


if __name__ == "__main__":
    unittest.main()



# This is a template for the "aminer" logdata-anomaly-miner tool. Copy
# it to "config.py" and define your ruleset. For more examples of component
# usage see aecid-testsuite/demo/aminer/demo-config.py.

config_properties = {}

# Define the list of log resources to read from: the resources
# named here do not need to exist when aminer is started. This
# will just result in a warning. However if they exist, they have
# to be readable by the aminer process! Supported types are:
# * file://[path]: Read data from file, reopen it after rollover
# * unix://[path]: Open the path as UNIX local socket for reading
config_properties['LogResourceList'] = ['file:///tmp/syslog']

# Define the uid/gid of the process that runs the calculation
# after opening the log files:
config_properties['AminerUser'] = 'aminer'
config_properties['AminerGroup'] = 'aminer'

learn_mode = True

# Read and store information to be used between multiple invocations
# of aminer in this directory. The directory must only be accessible
# to the 'AminerUser' but not group/world readable. On violation,
# aminer will refuse to start. When undefined, '/var/lib/aminer'
# is used.
# config_properties['Core.PersistenceDir'] = '/var/lib/aminer'

# Set the Unix-Domain-Socket for RemoteControl
# RemoteControlSocket: '/var/lib/aminer/log/remcontrol.sock'

# Add your ruleset here:


def build_analysis_pipeline(analysis_context):
    """
    Define the function to create pipeline for parsing the log data.
    It has also to define an AtomizerFactory to instruct aminer how to process incoming data streams to create log atoms from them.
    """
    # Build the parsing model:
    from aminer.parsing.SequenceModelElement import SequenceModelElement

    import ApacheAccessModel
    apache_access_model = ApacheAccessModel.get_model()

    parsing_model = SequenceModelElement('model', [apache_access_model])

    # Some generic imports.
    from aminer.analysis import AtomFilters

    # Create all global handler lists here and append the real handlers
    # later on.
    # Use this filter to distribute all atoms to the analysis handlers.
    atom_filter = AtomFilters.SubhandlerFilter(None)
    anomaly_event_handlers = []

    # Now define the AtomizerFactory using the model. A simple line
    # based one is usually sufficient.
    from aminer.input.SimpleByteStreamLineAtomizerFactory import SimpleByteStreamLineAtomizerFactory
    analysis_context.atomizer_factory = SimpleByteStreamLineAtomizerFactory(
        parsing_model, [atom_filter], anomaly_event_handlers, default_timestamp_path_list='/model/accesslog/time')

    # Just report all unparsed atoms to the event handlers.
    from aminer.analysis.UnparsedAtomHandlers import SimpleUnparsedAtomHandler
    atom_filter.add_handler(SimpleUnparsedAtomHandler(anomaly_event_handlers), stop_when_handled_flag=True)

    from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
    new_match_path_detector = NewMatchPathDetector(analysis_context.aminer_config, anomaly_event_handlers, learn_mode=learn_mode)
    analysis_context.register_component(new_match_path_detector, component_name=None)
    atom_filter.add_handler(new_match_path_detector)

    # Check if status-code changed
    from aminer.analysis.NewMatchPathValueDetector import NewMatchPathValueDetector
    new_match_path_value_detector = NewMatchPathValueDetector(
        analysis_context.aminer_config, ["/model/accesslog/status"], anomaly_event_handlers, learn_mode=learn_mode)
    analysis_context.register_component(new_match_path_value_detector, component_name=None)
    atom_filter.add_handler(new_match_path_value_detector)

    # Check if HTTP-Method for a HTTP-Request has changed
    from aminer.analysis.NewMatchPathValueComboDetector import NewMatchPathValueComboDetector
    new_match_path_value_combo_detector = NewMatchPathValueComboDetector(analysis_context.aminer_config, [
        "/model/accesslog/request", "/model/accesslog/method"], anomaly_event_handlers, learn_mode=learn_mode)
    analysis_context.register_component(new_match_path_value_combo_detector, component_name=None)
    atom_filter.add_handler(new_match_path_value_combo_detector)

    # Check if HTTP-Statuscode for a HTTP-Request has changed
    new_match_path_value_combo_detector2 = NewMatchPathValueComboDetector(analysis_context.aminer_config, [
        "/model/accesslog/request", "/model/accesslog/status"], anomaly_event_handlers, learn_mode=learn_mode)
    analysis_context.register_component(new_match_path_value_combo_detector2, component_name=None)
    atom_filter.add_handler(new_match_path_value_combo_detector2)

    # Add stdout stream printing for debugging, tuning.
    from aminer.events.StreamPrinterEventHandler import StreamPrinterEventHandler
    anomaly_event_handlers.append(StreamPrinterEventHandler(analysis_context))



"""This module defines an event handler that converts an event to JSON.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import json
import time
import copy

from aminer.events.EventInterfaces import EventHandlerInterface
from aminer import AminerConfig


class JsonConverterHandler(EventHandlerInterface):
    """This class implements an event record listener, that will convert event data to JSON format."""

    def __init__(self, json_event_handlers, analysis_context, pretty_print=True):
        """
        Initialize the event handler.
        @param json_event_handlers the event handlers to which the json converted data is sent.
        @param analysis_context the analysis context used to get the component.
        @param pretty_print if true, the json is printed pretty; otherwise the json is printed with less space needed.
        """
        self.json_event_handlers = json_event_handlers
        self.analysis_context = analysis_context
        self.pretty_print = pretty_print

    def receive_event(self, event_type, event_message, sorted_loglines, event_data, log_atom, event_source):
        """
        Receive information about a detected event.
        @param event_type is a string with the event type class this event belongs to. This information can be used to interpret
               type-specific event_data objects. Together with the eventMessage and sorted_loglines, this can be used to create generic log
               messages.
        @param event_message the first output line of the event.
        @param sorted_loglines sorted list of log lines that were considered when generating the event, as far as available to the time
               of the event. The list has to contain at least one line.
        @param event_data type-specific event data object, should not be used unless listener really knows about the event_type.
        @param log_atom the log atom which produced the event.
        @param event_source reference to detector generating the event.
        """
        if hasattr(event_source, 'output_event_handlers') and event_source.output_event_handlers is not None and self not in \
                event_source.output_event_handlers:
            return
        component_name = self.analysis_context.get_name_by_component(event_source)
        if component_name in self.analysis_context.suppress_detector_list:
            return
        if 'StatusInfo' in event_data:
            # No anomaly; do nothing on purpose
            pass
        else:
            log_data = {}
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            log_data['RawLogData'] = [data]
            if log_atom.get_timestamp() is None:
                log_atom.set_timestamp(time.time())
            log_data['Timestamps'] = [round(log_atom.atom_time, 2)]
            log_data['DetectionTimestamp'] = round(time.time(), 2)
            log_data['LogLinesCount'] = len(sorted_loglines)
            if log_atom.parser_match is not None and hasattr(event_source, 'output_logline') and event_source.output_logline:
                log_data['AnnotatedMatchElement'] = {}
                for path, match in log_atom.parser_match.get_match_dictionary().items():
                    if isinstance(match, list):
                        for match_element_id, match_element in enumerate(match):
                            if isinstance(match_element.match_object, bytes):
                                log_data['AnnotatedMatchElement'][path + '/' + str(match_element_id)] = match_element.match_object.decode(
                                    AminerConfig.ENCODING)
                            else:
                                log_data['AnnotatedMatchElement'][path + '/' + str(match_element_id)] = str(match_element.match_object)
                    elif isinstance(match.match_object, bytes):
                        log_data['AnnotatedMatchElement'][path] = match.match_object.decode(AminerConfig.ENCODING)
                    else:
                        log_data['AnnotatedMatchElement'][path] = str(match.match_object)

            analysis_component = {'AnalysisComponentIdentifier': self.analysis_context.get_id_by_component(event_source)}
            if event_source.__class__.__name__ == 'ExtractedData_class':
                analysis_component['AnalysisComponentType'] = 'DistributionDetector'
            else:
                analysis_component['AnalysisComponentType'] = str(event_source.__class__.__name__)
            analysis_component['AnalysisComponentName'] = self.analysis_context.get_name_by_component(event_source)
            analysis_component['Message'] = event_message
            if hasattr(event_source, "persistence_id"):
                analysis_component['PersistenceFileName'] = event_source.persistence_id
            if hasattr(event_source, 'learn_mode'):
                analysis_component['TrainingMode'] = event_source.learn_mode

            detector_analysis_component = event_data.get('AnalysisComponent')
            if detector_analysis_component is not None:
                for key in detector_analysis_component:
                    if key in analysis_component:
                        continue
                    analysis_component[key] = detector_analysis_component.get(key)

            if 'LogData' not in event_data:
                event_data['LogData'] = log_data
            event_data['AnalysisComponent'] = analysis_component

        if self.pretty_print is True:
            json_data = json.dumps(event_data, indent=2)
        else:
            json_data = json.dumps(event_data)
        res = [''] * len(sorted_loglines)
        res[0] = str(json_data)

        for listener in self.json_event_handlers:
            if hasattr(event_source, "output_event_handlers") and event_source.output_event_handlers is not None \
                    and listener not in event_source.output_event_handlers:
                event_source = copy.copy(event_source)
                event_source.output_event_handlers.append(listener)
            listener.receive_event(event_type, None, res, json_data, log_atom, event_source)



import unittest
from aminer.analysis.MatchValueAverageChangeDetector import MatchValueAverageChangeDetector
from aminer.parsing.MatchElement import MatchElement
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase
from datetime import datetime
import time


class MatchValueAverageChangeDetectorTest(TestBase):
    """Unittests for the MatchValueAverageChangeDetector."""

    __expected_string = '%s Statistical data report\n%s: "%s" (%d lines)\n  "cron/job1": Change: new: n = 3, avg = %s, ' \
                        'var = 100000000.0; old: n = 3, avg = %s, var = 1000000.0\n\n'
    __expected_string2 = '%s Statistical data report\n%s: "%s" (%d lines)\n  "cron/job1": Change: new: n = 2, avg = %s, ' \
                         'var = 50000000.0; old: n = 2, avg = %s, var = 500000.0\n  "cron/job2": Change: new: n = 2, avg = %s, ' \
                         'var = 60500000.0; old: n = 2, avg = %s, var = 500000.0\n\n'

    cron_job1 = 'cron/job1'
    cron_job2 = 'cron/job2'

    def test1receive_atom_min_bin_elements_not_reached(self):
        """This test verifies, that no statistic evaluation is performed, until the minimal amount of bin elements is reached."""
        description = "Test1MatchValueAverageChangeDetector"
        start_time = 57600
        match_element1 = MatchElement(self.cron_job1, b"%d" % start_time, start_time, None)
        match_value_average_change_detector = MatchValueAverageChangeDetector(self.aminer_config, [
            self.stream_printer_event_handler], None, [match_element1.get_path()], 3, start_time, False, 'Default')
        self.analysis_context.register_component(match_value_average_change_detector, description)

        # create oldBin
        log_atom = LogAtom(match_element1.get_match_object(), ParserMatch(match_element1), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 2000), start_time + 2000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 2000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # compare Data
        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 10000), start_time + 10000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 10000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 20000), start_time + 20000, None)
        log_atom = LogAtom(
          match_element1.get_match_object(), ParserMatch(match_element1), start_time + 20000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), '')

    def test2receive_atom_min_bin_time_not_reached(self):
        """This test verifies, that no statistic evaluation is performed, until the start time is reached."""
        description = "Test2MatchValueAverageChangeDetector"
        start_time = 57600
        match_element1 = MatchElement(self.cron_job1, b"%d" % start_time, start_time, None)
        match_value_average_change_detector = MatchValueAverageChangeDetector(
            self.aminer_config, [self.stream_printer_event_handler], 'time', [match_element1.get_path()], 3, start_time + 86400, False,
            'Default')
        self.analysis_context.register_component(match_value_average_change_detector, description)

        # create oldBin
        log_atom = LogAtom(match_element1.get_match_object(), ParserMatch(match_element1), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 2000), start_time + 2000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 2000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # compare Data
        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 10000), start_time + 10000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 10000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 20000), start_time + 20000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 20000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 30000), start_time + 30000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 30000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), '')

    def test3receive_atom_statistically_ok(self):
        """This test case focuses on receiving an atom and being in the statistically acceptable area."""
        description = "Test3MatchValueAverageChangeDetector"
        start_time = 57600
        match_element1 = MatchElement(self.cron_job1, b"%d" % start_time, start_time, None)
        match_value_average_change_detector = MatchValueAverageChangeDetector(
            self.aminer_config, [self.stream_printer_event_handler], 'time', [match_element1.get_path()], 3, start_time, False, 'Default')
        self.analysis_context.register_component(match_value_average_change_detector, description)

        # create oldBin
        log_atom = LogAtom(match_element1.get_match_object(), ParserMatch(match_element1), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 2000), start_time + 2000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 2000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # compare Data
        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 11000), start_time + 11000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 11000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 12000), start_time + 12000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 12000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 13000), start_time + 13000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 13000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), '')

    def test4receiveAtomStatisticallyOutOfRange(self):
        """This test case focuses on receiving an atom and being over the statistically acceptable area."""
        description = "Test4MatchValueAverageChangeDetector"
        start_time = time.time()

        match_element1 = MatchElement(self.cron_job1, b"%d" % start_time, start_time, None)
        match_value_average_change_detector = MatchValueAverageChangeDetector(
            self.aminer_config, [self.stream_printer_event_handler], None, [match_element1.get_path()], 3, start_time, False, 'Default')
        self.analysis_context.register_component(match_value_average_change_detector, description)

        # create oldBin
        log_atom = LogAtom(match_element1.get_match_object(), ParserMatch(match_element1), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 2000), start_time + 2000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 2000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # compare Data
        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 10000), start_time + 10000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 10000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 20000), start_time + 20000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 20000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 30000), start_time + 30000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 30000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(start_time + 30000).strftime("%Y-%m-%d %H:%M:%S"),
            match_value_average_change_detector.__class__.__name__, description, 6, start_time + 20000, start_time + 1000))

    def test5more_values(self):
        """This test case proves the functionality, when using more than one path."""
        description = "Test5MatchValueAverageChangeDetector"
        start_time = time.time()

        match_element1 = MatchElement(self.cron_job1, b"%d" % start_time, start_time, None)
        match_element2 = MatchElement(self.cron_job2, b"%d" % start_time, start_time, None)
        match_value_average_change_detector = MatchValueAverageChangeDetector(
            self.aminer_config, [self.stream_printer_event_handler], None, [
                match_element1.get_path(), match_element2.get_path()], 2, start_time, False, 'Default')
        self.analysis_context.register_component(match_value_average_change_detector, description)

        # create oldBin
        log_atom = LogAtom(match_element1.get_match_object(), ParserMatch(match_element1), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # create oldBin for ME2
        log_atom = LogAtom(match_element2.get_match_object(), ParserMatch(match_element2), start_time, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element2 = MatchElement(self.cron_job2, b"%d" % (start_time + 1000), start_time + 1000, None)
        log_atom = LogAtom(
            match_element2.get_match_object(), ParserMatch(match_element2), start_time + 1000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        # compare data
        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 10000), start_time + 10000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 10000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element1 = MatchElement(self.cron_job1, b"%d" % (start_time + 20000), start_time + 20000, None)
        log_atom = LogAtom(
            match_element1.get_match_object(), ParserMatch(match_element1), start_time + 20000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), '')

        # compare data with ME2
        match_element2 = MatchElement(self.cron_job2, b"%d" % (start_time + 11000), start_time + 11000, None)
        log_atom = LogAtom(
            match_element2.get_match_object(), ParserMatch(match_element2), start_time + 11000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)

        match_element2 = MatchElement(self.cron_job2, b"%d" % (start_time + 22000), start_time + 22000, None)
        log_atom = LogAtom(
            match_element2.get_match_object(), ParserMatch(match_element2), start_time + 22000, match_value_average_change_detector)
        match_value_average_change_detector.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string2 % (
            datetime.fromtimestamp(start_time + 22000).strftime("%Y-%m-%d %H:%M:%S"),
            match_value_average_change_detector.__class__.__name__, description, 4, start_time + 15000, start_time + 500,
            start_time + 16500, start_time + 500))


if __name__ == "__main__":
    unittest.main()



import unittest
import time
from unit.TestBase import TestBase
from aminer.analysis.MatchFilter import MatchFilter
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.ParserMatch import ParserMatch
from datetime import datetime


class MatchFilterTest(TestBase):
    """Unittests for the MatchFilter."""

    __expected_string = '%s Log Atom Filtered\nMatchFilter: "%s" (1 lines)\n  %d\n\n'

    def test1_receive_atom_trigger_event(self):
        """This test checks if an event is triggered if the path is in the target_path_list."""
        description = "Test1MatchFilterTest"
        decimal_integer_me = DecimalIntegerValueModelElement('integer')
        match_filter = MatchFilter(self.aminer_config, ['/integer'], [self.stream_printer_event_handler])
        self.analysis_context.register_component(match_filter, description)
        t = time.time()
        for val in range(1000):
            val_str = str(val).encode('utf-8')
            log_atom = LogAtom(val_str, ParserMatch(decimal_integer_me.get_match_element('', MatchContext(val_str))), t, match_filter)
            match_filter.receive_atom(log_atom)
            self.assertEqual(self.__expected_string % (
                datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S"), description, val), self.output_stream.getvalue())
            self.reset_output_stream()

    def test2_receive_atom_trigger_no_event(self):
        """This test checks if an event is not triggered if the path is not in the target_path_list."""
        description = "Test2MatchFilterTest"
        decimal_integer_me = DecimalIntegerValueModelElement('integer')
        match_filter = MatchFilter(self.aminer_config, ['/strings'], [self.stream_printer_event_handler])
        self.analysis_context.register_component(match_filter, description)
        t = time.time()
        for val in range(1000):
            val_str = str(val).encode('utf-8')
            log_atom = LogAtom(val_str, ParserMatch(decimal_integer_me.get_match_element('', MatchContext(val_str))), t, match_filter)
            match_filter.receive_atom(log_atom)
            self.assertEqual('', self.output_stream.getvalue())

    def test3_receive_atom_with_target_value(self):
        """This test checks if an event is triggered, when the path is in the target_path_list and the value is in the target_value_list."""
        description = "Test3MatchFilterTest"
        decimal_integer_me = DecimalIntegerValueModelElement('integer')
        match_filter = MatchFilter(self.aminer_config, ['/integer'], [self.stream_printer_event_handler], target_value_list=list(
            range(1001)))
        self.analysis_context.register_component(match_filter, description)
        t = time.time()
        for val in range(1000):
            val_str = str(val).encode('utf-8')
            log_atom = LogAtom(val_str, ParserMatch(decimal_integer_me.get_match_element('', MatchContext(val_str))), t, match_filter)
            match_filter.receive_atom(log_atom)
            self.assertEqual(self.__expected_string % (datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S"), description, val),
                             self.output_stream.getvalue())
            self.reset_output_stream()

    def test4_receive_atom_with_no_target_value(self):
        """
        This test checks if an event is not triggered.
        The path is in the target_path_list and the value is not in the target_value_list.
        """
        description = "Test4MatchFilterTest"
        decimal_integer_me = DecimalIntegerValueModelElement('integer')
        match_filter = MatchFilter(self.aminer_config, ['/integer'], [self.stream_printer_event_handler], target_value_list=list(
            range(501)))
        self.analysis_context.register_component(match_filter, description)
        t = time.time()
        for val in range(1000):
            val_str = str(val).encode('utf-8')
            log_atom = LogAtom(val_str, ParserMatch(decimal_integer_me.get_match_element('', MatchContext(val_str))), t, match_filter)
            match_filter.receive_atom(log_atom)
            if val <= 500:
                self.assertEqual(self.__expected_string % (datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S"), description, val),
                                 self.output_stream.getvalue())
            else:
                self.assertEqual('', self.output_stream.getvalue())
            self.reset_output_stream()


if __name__ == "__main__":
    unittest.main()



{
    'Analysis': {
        'required': False,
        'type': 'list',
        'nullable': True,
        'schema': {
            'type': 'dict',
            'schema': {
                'id': {'type': 'string', 'nullable': True, 'default': None},
                'type': {'type': 'analysistype', 'coerce': 'toanalysistype', 'required': True},
                'paths': {'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'labels': {'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'persistence_id': {'type': 'string', 'default': 'Default'},
                'output_logline': {'type': 'boolean', 'default': False},
                'learn_mode': {'type': 'boolean'},
                'num_windows': {'type': 'integer', 'required': True, 'default': 50},
                'min_anomaly_score': {'type': 'float', 'required': False, 'default': 1.1},
                'min_variance': {'type': 'float', 'required': False, 'default': 0.98},
                'allow_missing_values': {'type': 'boolean', 'default': False},
                'check_interval': {'type': 'integer', 'default': 3600},
                'realert_interval': {'type': 'integer', 'default': 36000},
                'report_interval': {'type': 'integer', 'default': 10},
                'reset_after_report_flag': {'type': 'boolean', 'default': False},
                'path': {'type': 'string', 'nullable': True, 'default': 'Default'},
                'parallel_check_count': {'type': 'integer', 'required': True, 'default': 10},
                'record_count_before_event': {'type': 'integer', 'default': 1000},
                'use_path_match': {'type': 'boolean', 'default': True},
                'use_value_match': {'type': 'boolean', 'default': True},
                'min_rule_attributes': {'type': 'integer', 'default': 1},
                'max_rule_attributes': {'type': 'integer', 'default': 5},
                'max_hypotheses': {'type': 'integer', 'default': 1000},
                'hypothesis_max_delta_time': {'type': 'float', 'default': 5.0},
                'generation_probability': {'type': 'float', 'default': 1.0},
                'generation_factor': {'type': 'float', 'default': 1.0},
                'max_observations': {'type': 'integer', 'default': 500},
                'p0': {'type': 'float', 'default': 0.9},
                'alpha': {'type': 'float', 'default': 0.05},
                'candidates_size': {'type': 'integer', 'default': 10},
                'hypotheses_eval_delta_time': {'type': 'float', 'default': 120.0},
                'delta_time_to_discard_hypothesis': {'type': 'float', 'default': 180.0},
                'check_rules_flag': {'type': 'boolean', 'default': True},
                'constraint_list': {
                    'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'ignore_list': {
                    'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'id_path_list': {'type': 'list', 'default': []},
                'scoring_path_list': {
                    'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'unique_path_list': {
                    'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'prob_thresh': {'type': 'float', 'default': 0.05},
                'default_freqs': {'type': 'boolean', 'default': False},
                'skip_repetitions': {'type': 'boolean', 'default': False},
                'seq_len': {'type': 'integer', 'default': 3},
                'timeout': {'type': ['integer', 'float'], 'default': -1},
                'allow_missing_id': {'type': 'boolean', 'default': False},
                'window_size': {'type': ['integer', 'float'], 'default': 600},
                'confidence_factor': {'type': 'float', 'default': 0.33},
                'min_allowed_time_diff': {'type': 'float', 'default': 5.0},
                'lower_limit': {'type': ['integer', 'float']},
                'upper_limit': {'type': ['integer', 'float']},
                'idf': {'type': 'boolean', 'default': False},
                'norm': {'type': 'boolean', 'default': False},
                'add_normal': {'type': 'boolean', 'default': False},
                'check_empty_windows': {'type': 'boolean', 'default': False},
                'bin_size': {'type': 'integer'},
                'bin_count': {'type': 'integer'},
                'outlier_bins_flag': {'type': 'boolean', 'default': False},
                'modulo_value': {'type': 'integer'},
                'time_unit': {'type': 'integer'},
                'histogram_defs': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string'}}},
                'bin_definition': {'type': 'string'},
                'tuple_transformation_function': {'type': 'string', 'allowed': ['demo'], 'nullable': True, 'default': None},
                'value_list': {
                    'type': 'list', 'schema': {'type': ['boolean', 'float', 'integer', 'string']}, 'nullable': True, 'default': None},
                'timestamp_path': {'type': 'string', 'nullable': True},
                'min_bin_elements': {'type': 'integer'},
                'min_bin_time': {'type': 'integer'},
                'debug_mode': {'type': 'boolean', 'default': False},
                # skipcq: PYL-W0511
                # TODO check which streams should be allowed
                'stream': {'type': 'string', 'allowed': ['sys.stdout', 'sys.stderr']},
                'separator': {'type': 'string'},
                'missing_value_string': {'type': 'string'},
                'subhandler_list': {'type': 'list', 'schema': {'type': 'string'}},
                'stop_when_handled_flag': {'type': 'boolean', 'default': False},
                'delete_components': {'type': 'boolean', 'default': True},
                'event_type': {'type': 'string'},
                'event_message': {'type': 'string'},
                'sub_rules': {'type': 'list', 'schema': {'type': 'string'}},
                'sub_rule': {'type': 'string'},
                'match_action': {'type': 'string', 'nullable': True, 'default': None},
                'rule_lookup_dict': {'type': 'dict'},
                'default_rule': {'type': 'string', 'nullable': True, 'default': None},
                'value': {'type': ['boolean', 'float', 'integer', 'string']},
                'regex': {'type': 'string'},
                'seconds_modulo': {'type': 'integer'},
                'limit_lookup_dict': {'type': 'dict', 'valuesrules': {'type': 'list', 'schema': {
                    'type': ['integer', 'float'], 'min': 0}}},
                'default_limit': {'type': 'list', 'schema': {'type': 'integer'}, 'nullable': True, 'default': None},
                'rule_id': {'type': 'string'},
                'min_time_delta': {'type': 'integer'},
                'max_time_delta': {'type': 'integer'},
                'artefact_match_parameters': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string'}},
                                              'nullable': True, 'default': None},
                'action_id': {'type': 'string'},
                'artefact_a_rules': {'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'artefact_b_rules': {'type': 'list', 'schema': {'type': 'string'}, 'nullable': True, 'default': None},
                'ruleset': {'type': 'list', 'schema': {'type': 'string'}},
                'exit_on_error_flag': {'type': 'boolean', 'default': False},
                'allowlist_rules': {'type': 'list', 'schema': {'type': 'string'}},
                'parsed_atom_handler_lookup_list': {
                    'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string', 'nullable': True}}},
                'default_parsed_atom_handler': {'type': 'string', 'nullable': True, 'default': None},
                'parsed_atom_handler_dict': {'type': 'dict', 'schema': {'id': {'type': 'string'}, 'type': {'type': 'string'}}},
                'min_num_vals': {'type': 'integer', 'default': 1000},
                'max_num_vals': {'type': 'integer', 'default': 1500},
                'save_values': {'type': 'boolean', 'default': True},
                'waiting_time': {'type': 'integer', 'default': 300},
                'num_sections_waiting_time': {'type': 'integer', 'default': 10},
                'event_type_detector': {'type': 'string'},
                'used_gof_test': {'type': 'string', 'allowed': ['CM', 'KS'], 'default': 'CM'},
                'gof_alpha': {'type': 'float', 'default': 0.05},
                's_gof_alpha': {'type': 'float', 'default': 0.05},
                's_gof_bt_alpha': {'type': 'float', 'default': 0.05},
                'd_alpha': {'type': 'float', 'default': 0.1},
                'd_bt_alpha': {'type': 'float', 'default': 0.1},
                'range_alpha': {'type': 'float', 'default': 0.05},
                'dw_alpha': {'type': 'float', 'default': 0.05},
                'div_thres': {'type': 'float', 'default': 0.3},
                'sim_thres': {'type': 'float', 'default': 0.1},
                'indicator_thres': {'type': 'float', 'default': 0.4},
                'num_init': {'type': 'integer', 'default': 100},
                'num_update': {'type': 'integer', 'default': 50},
                'num_update_unq': {'type': 'integer', 'default': 200},
                'num_s_gof_values': {'type': 'integer', 'default': 50},
                'num_s_gof_bt': {'type': 'integer', 'default': 30},
                'num_d_bt': {'type': 'integer', 'default': 30},
                'num_pause_discrete': {'type': 'integer', 'default': 5},
                'num_pause_others': {'type': 'integer', 'default': 2},
                'test_gof_int': {'type': 'boolean', 'default': True},
                'num_stop_update': {'type': 'boolean', 'default': False},
                'silence_output_without_confidence': {'type': 'boolean', 'default': False},
                'silence_output_except_indicator': {'type': 'boolean', 'default': True},
                'num_var_type_hist_ref': {'type': 'integer', 'default': 10},
                'num_update_var_type_hist_ref': {'type': 'integer', 'default': 10},
                'num_var_type_considered_ind': {'type': 'integer', 'default': 10},
                'num_stat_stop_update': {'type': 'integer', 'default': 200},
                'num_updates_until_var_reduction': {'type': 'integer', 'default': 20},
                'var_reduction_thres': {'type': 'float', 'default': 0.6},
                'num_skipped_ind_for_weights': {'type': 'integer', 'default': 1},
                'num_ind_for_weights': {'type': 'integer', 'default': 100},
                'used_multinomial_test': {'type': 'string', 'allowed': ['Approx', 'MT', 'Chi'], 'default': 'Chi'},
                'use_empiric_distr': {'type': 'boolean', 'default': True},
                'save_statistics': {'type': 'boolean', 'default': True},
                'split_reports_flag': {'type': 'boolean', 'default': False},
                'disc_div_thres': {'type': 'float', 'default': 0.3},
                'num_steps_create_new_rules': {'type': 'integer', 'default': -1},
                'num_upd_until_validation': {'type': 'integer', 'default': 20},
                'num_end_learning_phase': {'type': 'integer', 'default': -1},
                'check_cor_thres': {'type': 'float', 'default': 0.5},
                'check_cor_prob_thres': {'type': 'float', 'default': 1.0},
                'check_cor_num_thres': {'type': 'integer', 'default': 10},
                'min_values_cors_thres': {'type': 'integer', 'default': 5},
                'new_vals_alarm_thres': {'type': 'float', 'default': 3.5},
                'num_bt': {'type': 'integer', 'default': 30},
                'alpha_bt': {'type': 'float', 'default': 0.1},
                'used_homogeneity_test': {'type': 'string', 'allowed': ['Chi', 'MaxDist'], 'default': 'Chi'},
                'used_range_test': {'type': 'string', 'allowed': ['MeanSD', 'EmpiricQuantiles', 'MinMax'], 'default': 'MinMax'},
                'range_threshold': {'type': 'float', 'default': 1},
                'range_limits_factor': {'type': 'float', 'default': 1},
                'num_reinit_range': {'type': 'integer', 'default': 100},
                'alpha_chisquare_test': {'type': 'float', 'default': 0.05},
                'max_dist_rule_distr': {'type': 'float', 'default': 0.1},
                'used_presel_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': [
                    'matchDiscDistr', 'excludeDueDistr', 'matchDiscVals', 'random']}, 'nullable': True, 'default': None},
                'intersect_presel_meth': {'type': 'boolean', 'default': False},
                'percentage_random_cors': {'type': 'float', 'default': 0.20},
                'match_disc_vals_sim_tresh': {'type': 'float', 'default': 0.7},
                'exclude_due_distr_lower_limit': {'type': 'float', 'default': 0.4},
                'match_disc_distr_threshold': {'type': 'float', 'default': 0.5},
                'used_cor_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': ['Rel', 'WRel']},
                                  'nullable': True, 'default': None},
                'used_validate_cor_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': [
                    'coverVals', 'distinctDistr']}, 'nullable': True, 'default': None},
                'validate_cor_cover_vals_thres': {'type': 'float', 'default': 0.7},
                'validate_cor_distinct_thres': {'type': 'float', 'default': 0.05},
                'time_period_length': {'type': 'integer', 'min': 1, 'required': True, 'default': 86400},
                'max_time_diff': {'type': 'integer', 'min': 1, 'required': True, 'default': 360},
                'num_reduce_time_list': {'type': 'integer', 'min': 1, 'required': True, 'default': 10},
                'output_event_handlers': {'type': 'list', 'nullable': True, 'default': None},
                'suppress': {'type': 'boolean', 'default': False},
                'build_sum_over_values': {'type': 'boolean', 'default': False},
                'num_division_time_step': {'type': 'integer', 'default': 10},
                'num_min_time_history': {'type': 'integer', 'default': 20},
                'num_max_time_history': {'type': 'integer', 'default': 30},
                'num_results_bt': {'type': 'integer', 'default': 15},
                'round_time_interval_threshold': {'type': 'float', 'default': 0.02},
                'acf_threshold': {'type': 'float', 'default': 0.2},
                'acf_pause_interval_percentage': {'type': 'float', 'default': 0.2},
                'acf_auto_pause_interval': {'type': 'boolean', 'default': True},
                'acf_auto_pause_interval_num_min': {'type': 'integer', 'min': 1, 'required': True, 'default': 10},
                'num_log_lines_solidify_matrix': {'type': 'integer', 'default': 10000},
                'time_output_threshold': {'type': 'integer', 'default': 0},
                'anomaly_threshold': {'type': 'float', 'default': 0.05},
                'num_periods_tsa_ini': {'type': 'integer', 'default': 20},
                'allowed_id_tuples': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string'}}, 'nullable': True,
                                      'default': None},
                'force_period_length': {'type': 'boolean', 'default': False},
                'set_period_length': {'type': 'integer', 'default': 604800},
                'min_log_lines_per_time_step': {'type': 'integer', 'default': 10},
                'empty_window_warnings': {'type': 'boolean', 'default': True},
                'early_exceeding_anomaly_output': {'type': 'boolean', 'default': False},
                'set_lower_limit': {'type': 'integer', 'min': 0, 'nullable': True, 'default': None},
                'set_upper_limit': {'type': 'integer', 'min': 0, 'nullable': True, 'default': None},
                'local_maximum_threshold': {'type': 'float', 'default': 0.2},
                'combine_values': {'type': 'boolean', 'nullable': True, 'default': True},
            }
        }
    }
}



"""
This module defines an event handler that adds a confidence score to the anomaly output.
The score is calculated through analysis of a list of strings defined in the detector through the function get_weight_analysis_field_path
and weights the single strings based on the weights dictionary.
The weights can optionally be automatically calculated.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import copy

from aminer.events.EventInterfaces import EventHandlerInterface
from aminer.events.EventInterfaces import EventSourceInterface


class ScoringEventHandler(EventHandlerInterface):
    """This class implements an event record listener, that will add a confidence score to the anomaly output."""

    def __init__(self, event_handlers, analysis_context, weights=None, auto_weights=False, auto_weights_history_length=1000):
        """
        Initialize the ScoringEventHandler component.
        @param weights A dictionary that specifies the weights of values for the scoring. The keys are the strings of the analyzed list and
        the corresponding values are the assigned weights. Strings that are not present in this dictionary have the weight 0.5 if not
        automatically weighted.
        @param auto_weights boolean value that states if the weights should be automatically calculated through the formula
        10 / (10 + number of value appearances).
        @param auto_weights_history_length integer value that specifies the number of values that are considered in the calculation of the
        weights.
        """
        self.analysis_context = analysis_context
        self.event_handlers = event_handlers
        self.weights = weights
        self.auto_weights = auto_weights
        self.auto_weights_history_length = auto_weights_history_length

        if self.auto_weights:
            self.history_list = [[] for _ in range(self.auto_weights_history_length)]
            self.history_list_index = 0

    def receive_event(self, event_type, event_message, sorted_log_lines, event_data, log_atom, event_source):
        """Receive information about a detected event."""
        # Initialize path_valid variable that states if the path to the analysis field is valid
        path_valid = True

        # Get the path to the analysis and output fields from the event_source or set the paths to empty lists if not
        if isinstance(event_source, EventSourceInterface):
            analysis_field_path = event_source.get_weight_analysis_field_path()
            output_field_path = event_source.get_weight_output_field_path()
        else:
            analysis_field_path = []
            output_field_path = []

        # Check if the analysis field path is not empty and get the analyis list or set path_valid to False
        if analysis_field_path == []:
            path_valid = False
        else:
            analyis_list = event_data
            # Traverse the path of the analysis_field_path in event_data
            for path in analysis_field_path:
                if path in analyis_list:
                    # Go a step in the event_data
                    analyis_list = analyis_list[path]
                else:
                    # Set path_valid to False and stop if the path does not match the structure of event_data
                    path_valid = False
                    break

        # Calculate and add the confidence to the output if the path is valid
        if path_valid:
            event_data_confidence = event_data
            # Traverse the path of the output_field_path in event_data
            for path in output_field_path[:-1]:
                # Create a new dictionary if the path does not exist
                if path not in event_data_confidence:
                    event_data_confidence[path] = {}
                # Go a step in the event_data
                event_data_confidence = event_data_confidence[path]

            # Calculate the absolute confidence
            confidence_absolut = sum([self.get_weight(val) for val in analyis_list])
            # Add the the absolute and mean confidence to the message
            event_data_confidence[output_field_path[-1]] = {'confidence_absolut': confidence_absolut,
                                                            'confidence_mean': confidence_absolut / len(analyis_list)}

            # Update the history list and increase the count index
            if self.auto_weights:
                self.history_list[self.history_list_index] = analyis_list
                self.history_list_index += 1
                if self.history_list_index >= self.auto_weights_history_length:
                    self.history_list_index %= self.auto_weights_history_length

        # Send the message to the following event handlers
        for listener in self.event_handlers:
            if hasattr(event_source, "output_event_handlers") and event_source.output_event_handlers is not None \
                    and listener not in event_source.output_event_handlers:
                event_source = copy.copy(event_source)
                event_source.output_event_handlers.append(listener)
            listener.receive_event(event_type, event_message, sorted_log_lines, event_data, log_atom, event_source)

    def get_weight(self, value):
        """Return the weight of the value parameter."""
        if self.weights is not None and value in self.weights:
            # Return the specified weight if the value is in the weight list
            return self.weights[value]
        if not self.auto_weights:
            # Return 0.5 if the value is not in the weigth list and the weights are not automatically calculated
            return 0.5
        # Else calculate the weight through 10 / (10 + number of value appearances)
        return 10 / (10 + sum([value in value_list for value_list in self.history_list]))



import unittest
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase


class MatchElementTest(TestBase):
    """Unittests for the MatchElement."""

    path = "path"
    match_string = b"12.5"
    match_object = 12.5

    def test1get_path(self):
        """Test if get_path works properly."""
        match_element = MatchElement(self.path, self.match_string, self.match_object, None)
        self.assertEqual(match_element.get_path(), self.path)

    def test2get_match_string(self):
        """Test if get_match_string returns None."""
        match_element = MatchElement(self.path, self.match_string, self.match_object, None)
        self.assertEqual(match_element.get_match_string(), self.match_string)

    def test3get_match_object(self):
        """Test if get_match_string returns None."""
        match_element = MatchElement(self.path, self.match_string, self.match_object, None)
        self.assertEqual(match_element.get_match_object(), self.match_object)

    def test4get_children(self):
        """Test if get_match_string returns None."""
        match_element = MatchElement(self.path, self.match_string, self.match_object, None)
        self.assertEqual(match_element.get_children(), None)

    def test5annotate_match(self):
        """This test case checks if all possible annotations are created correctly."""
        a3 = MatchElement("a3", b"a3", b"a3", None)
        a2 = MatchElement("a2", b"a2", b"a2", [a3])
        a1 = MatchElement("a1", b"a1", b"a1", [a2])
        b3 = MatchElement("b3", b"b3", b"b3", None)
        b2 = MatchElement("b2", b"b2", b"b2", [b3])
        b1 = MatchElement("b1", b"b1", b"b1", [b2])
        root_element = MatchElement("root", b"root", b"root", [a1, b1])

        self.assertEqual(root_element.annotate_match(None), "root: root a1: a1 a2: a2 a3: a3 b1: b1 b2: b2 b3: b3")
        self.assertEqual(root_element.annotate_match(""), "root: root\n  a1: a1\n    a2: a2\n      a3: a3\n  b1: b1\n    b2: b2\n      "
                                                          "b3: b3")
        self.assertEqual(root_element.annotate_match("--"), "--root: root\n--  a1: a1\n--    a2: a2\n--      a3: a3\n--  b1: b1\n"
                                                            "--    b2: b2\n--      b3: b3")

    def test6serialize_object(self):
        """This test case checks if all child objects are serialized correctly."""
        a3 = MatchElement("a3", b"a3", b"a3", None)
        a2 = MatchElement("a2", b"a2", b"a2", [a3])
        a1 = MatchElement("a1", b"a1", b"a1", [a2])
        b3 = MatchElement("b3", b"b3", b"b3", None)
        b2 = MatchElement("b2", b"b2", b"b2", [b3])
        b1 = MatchElement("b1", b"b1", b"b1", [b2])
        root_element = MatchElement("root", b"root", b"root", [a1, b1])

        self.assertEqual(root_element.serialize_object(), {"path": "root", "match_object": b"root", "match_string": b"root", "children": [
            {"path": "a1", "match_object": b"a1", "match_string": b"a1", "children": [
                {"path": "a2", "match_object": b"a2", "match_string": b"a2",
                    "children": [{"path": "a3", "match_object": b"a3", "match_string": b"a3", "children": []}]}]},
            {"path": "b1", "match_object": b"b1", "match_string": b"b1", "children": [
                {"path": "b2", "match_object": b"b2", "match_string": b"b2",
                    "children": [{"path": "b3", "match_object": b"b3", "match_string": b"b3", "children": []}]}]}]})

    def test7str(self):
        """Test the string representation of the MatchElements."""
        a3 = MatchElement("a3", b"a3", b"a3", None)
        a2 = MatchElement("a2", b"a2", b"a2", [a3])
        a1 = MatchElement("a1", b"a1", b"a1", [a2])
        b3 = MatchElement("b3", b"b3", b"b3", None)
        b2 = MatchElement("b2", b"b2", b"b2", [b3])
        b1 = MatchElement("b1", b"b1", b"b1", [b2])
        root_element = MatchElement("root", b"root", b"root", [a1, b1])
        self.assertEqual(root_element.__str__(), "MatchElement: path = root, string = root, object = root, children = 2")

        root_element = MatchElement("match", b"string", 2, None)
        self.assertEqual(root_element.__str__(), "MatchElement: path = match, string = string, object = 2, children = 0")

    def test8init_path_input_validation(self):
        """Check if path is validated in __init__()."""
        self.assertRaises(TypeError, MatchElement, b"path", self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, True, self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, 123, self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, 123.22, self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, {"id": "path"}, self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, ["path"], self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, [], self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, (), self.match_string, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, set(), self.match_string, self.match_object, None)

    def test9init_match_string_input_validation(self):
        """Check if match_string is validated in __init__()."""
        self.assertRaises(TypeError, MatchElement, self.path, "path", self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, True, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, 123, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, 123.22, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, {"id": "path"}, self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, ["path"], self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, [], self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, (), self.match_object, None)
        self.assertRaises(TypeError, MatchElement, self.path, set(), self.match_object, None)

    def test10init_match_object_input_validation(self):
        """Check if match_object is validated in __init__()."""
        MatchElement(self.path, self.match_string, b"", None)
        MatchElement(self.path, self.match_string, "path", None)
        MatchElement(self.path, self.match_string, True, None)
        MatchElement(self.path, self.match_string, 123, None)
        MatchElement(self.path, self.match_string, 123.22, None)
        MatchElement(self.path, self.match_string, {"id": "path"}, None)
        MatchElement(self.path, self.match_string, ["path"], None)
        MatchElement(self.path, self.match_string, [], None)
        MatchElement(self.path, self.match_string, (), None)
        MatchElement(self.path, self.match_string, set(), None)
        MatchElement(self.path, self.match_string, MatchElement(self.path, self.match_string, self.match_object, None), None)

    def test11init_children_input_validation(self):
        """Check if children is validated in __init__()."""
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, b"path")
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, "path")
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, True)
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, 123)
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, 123.22)
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, {"id": "path"})
        self.assertRaises(ValueError, MatchElement, self.path, self.match_string, self.match_object, [])
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, ())
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, set())
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, ["string"])
        self.assertRaises(TypeError, MatchElement, self.path, self.match_string, self.match_object, [b"string"])

    def test12init_child_elements_with_no_path(self):
        """This test case checks, whether an exception is raised, when the path is None and children are passed."""
        self.assertRaises(ValueError, MatchElement, None, self.match_string, self.match_object, [
            MatchElement(self.path, self.match_string, self.match_object, None)])

    def test13annotate_match_indent_str_input_validation(self):
        """Check if indent_str is validated in annotate_match()."""
        match_element = MatchElement(self.path, self.match_string, self.match_object, None)
        self.assertRaises(TypeError, match_element.annotate_match, b"  ")
        self.assertRaises(TypeError, match_element.annotate_match, [" ", "-"])
        self.assertRaises(TypeError, match_element.annotate_match, 123.22)
        self.assertRaises(TypeError, match_element.annotate_match, {"id": "path"})
        self.assertRaises(TypeError, match_element.annotate_match, ["path"])
        self.assertRaises(TypeError, match_element.annotate_match, [])
        self.assertRaises(TypeError, match_element.annotate_match, ())
        self.assertRaises(TypeError, match_element.annotate_match, set())


if __name__ == "__main__":
    unittest.main()
